/** \file server.c
 * \brief TINE server engine plus API routines
 *
 */
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "srvcore.h"
#include "threader.h"
#include "globals.h"
#include "errors.h"
#include "dbtools.h"
#include "almlib.h"
#include "hstlib.h"
#include "srvip.h"
#include "srvipx.h"
#include "ttycmd.h"
#if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
# include "shmalloc.h"
#endif

int canSendGlobals = CAN_SENDGLOBALS;
int ngcasts = 0, nglobals = 0;
int gBurstLimit = BURST_LIMIT;
int gCycleDelay = CYCLE_DELAY;
TINE_EXPORT int SystemPresetMemory = TRUE;
int gDelayInitTineServices = FALSE;
int gSystemUseDataProtection = FALSE;
int gInitTineServicesDone = FALSE;
extern int gInitIPServicesDone;
extern int gInitIPClientDone;
extern int history_home_done;

TINE_EXPORT void SetSystemUseDataProtection(int value)
{
  gSystemUseDataProtection = value;
}
#ifndef srvWorkArea
BYTE *srvWorkArea;
UINT32 srvWorkAreaSize = WORK_SIZE;
#else
BYTE *WorkArea;
UINT32 WorkAreaSize = WORK_SIZE;
#endif
BYTE *tmpWorkArea = NULL;
UINT32 tmpWorkAreaSize = WORK_SIZE;

void SurveyAlarms(struct timeval *tv);
void doGCast(struct timeval *thisTime);
void DataDump(char *txt,char *data,int siz,short fmt,short dir);
int datacmp(BYTE *d1,BYTE *d2,int siz,int fmt,double t);
void CloseLink(int i);
void CleanupStrandedDatagrams(void);
void _SystemEngine(void);
int initTineServices(void);
void reAttachToServiceGroup(void);

int gEqmFcnAccessor = 0;
char *SRVDataBuffer;          /* global Buffer pointers for data exchange */
UINT32 SysPoll = DEFAULT_POLLRATE;
TINE_EXPORT UINT16 MinPollingRate = MIN_POLLRATE;
TINE_EXPORT UINT16 MaxPollingRate = DEFAULT_POLLRATE;
TINE_EXPORT UINT32 MaxSystemTransportSize = MAX_TRANSPORTSIZE;
int gBrdcstBufferSize = BRDCST_BUFFER_SIZE; 
int AveBusyTime = 0;
struct timeval thisCycleTime;
int RequireAcknowledgments = 1; /* global flag to force acknowledge in REFRESH mode */
short gNumEqmBkgTasks = 0;
int nExpiredContracts = 0;

AccessLockStruct currentAccessLock;
GlobalListStruct **gcastTbl = NULL;
IPXAddress IPXFecAddress;
int ExportsRegistered = 0;
int FecNameRegistered = 0;
FecInfoStruct gFecInfo;

int gSystemDataStamp = 0;
TINE_EXPORT void SetSystemDataStamp(int value)
{
  gSystemDataStamp = value;
}
int SystemRunning = FALSE;
TINE_EXPORT int IsServerRunning(void)
{
  return SystemRunning;
}
TINE_EXPORT int gRetardSingleContractRemoval = TRUE;
SINT32 findpoll(SINT32 p1,SINT32 p2)
{
  SINT32 p3;

  if (MinPollingRate > MaxPollingRate) MinPollingRate = MaxPollingRate;
  if (p1 < MinPollingRate) return p2 <= MinPollingRate ? MinPollingRate : p2;
  if (p2 < MinPollingRate || p1 == p2) return p1;
  if (p1 <= 1) return p1; if (p2 <= 1) return p2;
  if (p2 < p1) while ((p3 = p1%p2) >= MinPollingRate) { p1 = p2; p2=p3; }
  else if (p1 < p2) while ((p3 = p2%p1) >= MinPollingRate) { p2 = p1; p1=p3; }
  return p1 > p2 ? p2 : p1;
}
ExportListStruct *ExportList = NULL;
#ifdef MULTITHREADED
int gNeedToStartEqmThreads = FALSE;
void StartEqmThreads(void)
{
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->hEqmCallThread != 0) continue;
    if (useMultiThreadedEquipmentFunctions) 
    { /* launch the call handler thread */
      el->hEqmCallThread = eqmCreateThread((THREAD_PARAM)el);
      feclog("FEC: Equipment Function %s managed in separate thread",el->EqmName);
    }
  }
  gNeedToStartEqmThreads = FALSE;
  return;
}
#endif
void InitEqpModules(void)
{
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->inidone) continue;
    if (el->eqmIni) el->eqmIni();
    historyInit(el->EqmExportName,el->EqmName);
    el->inidone = TRUE;
  }
# ifdef MULTITHREADED
  if (useMultiThreadedEquipmentFunctions) gNeedToStartEqmThreads = TRUE;
# endif
  return;
}
void devNamesInitialization(void)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int plen, idx;
  char prpName[64];

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->hasPropertyDeviceNames)
    {
      for (idx=0; idx<PROPERTY_HASH_SIZE; idx++)
        for (prp=el->EqmProperty[idx]; prp; prp=prp->next)
      {
        if ((plen=(int)strlen(prp->prpName)) < 5) continue;
        if (strcmp(&prp->prpName[plen-4],".NAM")) continue;
        strncpy(prpName,prp->prpName,plen-4); prpName[plen-4] = 0;
        AssignDeviceListFromProperty(el->EqmName,prpName);
      }
    }
  }
  acquireDeviceNamesFromProperties = FALSE;
}
int LastMinuteInitialization(void)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int idx;

  InitEqpModules();
# ifdef MULTITHREADED
  if (useMultiThreadedStockFunctions) 
  {
    hStkThread = stkCreateThread(0);
    feclog("FEC: Stock functions managed in separate thread");
  }
# endif

  if (gDelayInitTineServices) initTineServices();

  if (StartupDebug)
  {
    printf("Exporting the following properties:\n>");
    for (el=ExportList; el != NULL; el=el->next)
      for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
        for (prp=el->EqmProperty[idx]; prp; prp=prp->next)
    {
      printf("%s (%s):  %s (%s) -> %s <-\n>",el->EqmExportName,
             el->EqmName,prp->prpName,
             itom(prp->prpAccessMode),prp->prpDescription);
    }
  }
  return 1;  /* init done */
}
int RegisterLocalServiceModule(void)
{
  ExportListStruct *el;
  int cc;
  if ((cc=RegisterEquipmentModule(SRVEXP_NAME,SRVEQM_NAME,0,NULL,NULL,NULL,0,NULL)) != 0) return cc;
  if ((el=getExportListItem(SRVEQM_NAME)) == NULL) return non_existent_elem;
  el->registered = REGISTERED_ALL; /* don't allow this to be sent to the ENS ! */
  return 0;
}
void ExitEQPModules(void)
{
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next)
  {
#   ifdef MULTITHREADED
    if (el->hEqmBkg != 0) threadExit(el->hEqmBkg);
    if (el->hEqmCallThread != 0) threadExit(el->hEqmCallThread);
#   endif
    if (el->eqmExi) el->eqmExi();  /* call exit routines */
  }
  return;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT void FreeEQPModules(void)
{
  ExportListStruct *el, *nxtel=NULL;
  ExportPropertyListStruct *prp=NULL, *nxtp=NULL;
  HashXRefStruct *dev=NULL,*nxtd=NULL;
  ALARM *alm=NULL, *nxta=NULL;
  ADSList *lst=NULL, *nxtl=NULL;
  AWSList *wlst=NULL, *nxtw=NULL;
  int i, k;

  ExitEQPModules();

  ServerExitCondition = 1;
  _SystemReset(0);
  for (el=ExportList; el != NULL; )
  {
    feclog("freeing equipment module memory for %s",el->EqmName);
    for (i=0; i<PROPERTY_HASH_SIZE+1; i++)
    {
      for (prp=el->EqmProperty[i]; prp != NULL; )
      {
        nxtp = prp->next;
        SystemFree(prp);
        prp = nxtp;
      }
    }
    if (el->EqmDeviceXRef != NULL)
    {
      for (i=0; i<DEVICES_HASH_SIZE; i++)
      {
        for (dev=el->EqmDeviceXRef[i]; dev != NULL; )
        {
          nxtd = dev->next;
          SystemFree(dev);
          dev = nxtd;
        }
      }
    }
    if (el->EqmDeviceList != NULL) SystemFree(el->EqmDeviceList);
    if (el->EqmDevicePrpList != NULL) 
    {
      for (i=0; i<el->EqmDeviceSpaceSize; i++)
      {
        if (el->EqmDevicePrpList[i].prpList != NULL) 
        {
          SystemFree(el->EqmDevicePrpList[i].prpList);
          for (k=i+1; k<el->EqmDeviceSpaceSize; k++)
          { /* remove dependencies */
            if (el->EqmDevicePrpList[k].prpList == el->EqmDevicePrpList[i].prpList)
            {
              el->EqmDevicePrpList[k].prpList = NULL;
            }
          }
        }
        if (el->EqmDeviceRdrList != NULL && el->EqmDeviceRdrList[i].rdrName != NULL)
        {
          SystemFree(el->EqmDeviceRdrList[i].rdrName);
        }
      }
    }
    if (el->EqmDeviceRdrList != NULL) SystemFree(el->EqmDeviceRdrList);
    if (el->adsTable != NULL) 
    {
      for (lst=el->adsTable; lst !=NULL; )
      {
        nxtl = lst->next;
        SystemFree(lst);
        lst = nxtl;
      }
    }
    if (el->almLst != NULL) 
    {
      for (i=0; i<el->EqmDeviceSpaceSize; i++)
      {
        for (alm=el->almLst[i]; alm !=NULL; )
        {
          nxta = alm->next;
          SystemFree(alm);
          alm = nxta;
        }
      }
      SystemFree(el->almLst);
    }
    nxtel = el->next;
    SystemFree(el);
    el = nxtel;
  }
  ExportList = NULL;
  if (hstTbl != NULL)
  {
    feclog("freeing local history table (%d records)",nHistoryRecords);
    for (i=0; i<maxHistoryRecords; i++)
    {
      if (hstTbl[i].prd != NULL) SystemFree(hstTbl[i].prd);
      if (hstTbl[i].timestamp != NULL) SystemFree(hstTbl[i].timestamp);
      if (hstTbl[i].data != NULL) SystemFree(hstTbl[i].data);
    }
    SystemFree(hstTbl);
  }
  nHistoryRecords = 0;
  hstTbl = NULL;
  history_home_done = FALSE;
  if (almWatchList != NULL)
  {
    feclog("freeing alarm watch table (%d records)",nAlmWatchList);
    for (wlst=almWatchList; wlst !=NULL; )
    {
      nxtw = wlst->nxt;
      SystemFree(wlst);
      wlst = nxtw;
    }
  }
  almWatchList = NULL;
  nAlmWatchList = 0;
  return;
}
#endif /* SMALL_TINE_LIB */
void RegisterEQPtasks(char *name,void (*ini)(void),void (*tsk)(void),int rate,void (*exi)(void))
{
  ExportListStruct *el;

  if ((el=getExportListItem(name)) == NULL) return;

  if ((el->eqmBkg=tsk) != NULL) gNumEqmBkgTasks++;
  el->rateBkg = rate;
  if (ServerCycleMode != POLLING && rate < MaxPollingRate)
  {
    SysPoll = rate;
    MaxPollingRate = rate;
    if (rate < MinPollingRate) MinPollingRate = rate;
  }
  el->eqmIni = ini;
  el->eqmExi = exi;

  if (StartupDebug)
    printf("%12s task registration: %s %s %s\n>",name,
            ini ? "INIT" : "----",
            tsk ? "BCKG" : "----",
            exi ? "EXIT" : "----");

  return;
}
void checkTasklist(void)
{
  SINT32 msec;
  struct timeval tv;
  ExportListStruct *el;

  gettimeofday(&tv,(struct timezone *)NULL);

  for (el=ExportList; el != NULL; el=el->next) 
  {
    if (el->eqmBkg)
    {
      if ((msec=MSECS(tv,el->lastBkg)) < el->rateBkg) continue;
      if (msec == 0 && el->rateBkg != 0) continue;
      if (el->bkgbusy) continue;
      el->bkgbusy = 1;
      el->lastBkg = tv;
      el->eqmBkg();
      el->bkgbusy = 0;
      if (el->next != NULL) _SystemEngine();
    }
  }
  return;
}
int WRITEaccessFromAllNets = 0; /* leave this undocumented! */
short IsIllegalProperty(char *eqmName,char *prpName,char *dev,CLN *cln,short access,ExportListStruct **listEntry,char *stsString)
{
  int k,plen=0,ps;
  short prpType,cc=0,rc;
  ExportPropertyListStruct *prp=NULL;
  ExportListStruct *el;
  char *c,Property[PROPERTY_NAME_SIZE];

  stsString[0] = 0;
  strncpy(Property,prpName,PROPERTY_NAME_SIZE);
  for (k=0; k<NrStockModifiers; k++)
  { /* meta property request? -> prepare base property */
    if ((c=strstr(Property,StockModifier[k].modName)) != NULL)
    {
      *c = 0;
      break;
    }
  }
  prpType = EMPTY_PROP;
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqmName,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL)
  { /* came up emty, try the alias name */
    for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqmName,el->EqmAlias,EQM_NAME_SHORTSIZE)) break;
  }
  if (el == NULL) ccerr(non_existent_elem);
  if (el->EqmPropertyQueryFcn == NULL && el->XEqmPropertyQueryFcn == NULL) 
  { /* use default property list */
    if ((rc=lookupPropertyName(el,prpName,dev)) == -server_redirection) ccerr(-rc);
    if (rc < 0 && (rc=lookupPropertyName(el,Property,dev)) < 0) 
    { /* failed first test */
      if (rc == -server_redirection) ccerr(-rc);
      /* check for wild cards */
      for (prp=el->EqmProperty[PROPERTY_HASH_SIZE]; prp; prp=prp->next)
      {
        for (ps = 0; ps < (int)strlen(prp->prpName) && prp->prpName[ps] == '*'; ps++);
        plen = (ps < (int)strlen(prp->prpName)) ? (int)strcspn(&prp->prpName[ps],"*") : 0;
        if (plen && !strnicmp(&Property[ps],&prp->prpName[ps],plen)) break;
      }
      if (prp == NULL) 
      { /* Maybe it's a query */
        if ((prpType=IsStockProperty(Property)) != STOCK_PROP) ccerr(illegal_property);
      }
      else prpType = WILDCARD_PROP;
    }
    else 
    { /*  is registered normally */
      prpType = TARGET_PROP;
    }
  }
  else  
  { /* uses a property query function, but check a few things up front ... */
    if (lookupPropertyName(el,prpName,dev) == -server_redirection) ccerr(server_redirection);
    if (strnicmp(prpName,Property,PROPERTY_NAME_SIZE) && lookupPropertyName(el,Property,dev) == -server_redirection) 
      ccerr(server_redirection);
    if (lookupRedirectionNameStub != NULL) 
    {
      if (lookupRedirectionNameStub(el->EqmName,prpName,dev) == server_redirection) ccerr(server_redirection);
      if (strnicmp(prpName,Property,PROPERTY_NAME_SIZE) && 
          lookupRedirectionNameStub(el->EqmName,Property,dev) == server_redirection) ccerr(server_redirection);
    }
    prpType = IsStockProperty(Property);
  }
  if ((access&CA_WRITE) && strcmp("MESSAGE",Property) && cln != NULL) /* not a MESSAGE */
  {
    for (k=0;!WRITEaccess && k<NrOfUsers;k++)
    {
      if (!strnicmp(gAllowedUsersList[k].name,cln->userName,USERNAME_SIZE)) break;
    }
    if (!WRITEaccess && k == NrOfUsers) 
    {
      feclog("COMMAND %s called by %s : user refused access",Property,cln->userName);
      ccerr(access_denied);
    }
    if (!WRITEaccessFromAllNets &&
        !IsMemberControlNets(cln)) 
    {
      inet_ntoa_b(cln->IPaddress.sin_addr,gClientAddrBuffer);
      feclog("COMMAND %s called by %s from addr %s: network refused access",
        Property,cln->userName,gClientAddrBuffer);
      ccerr(access_denied);
    }
    if (strcmp("ACCESSLOCK",Property) && !IsMemberLockSet(cln))
    {
      inet_ntoa_b(cln->IPaddress.sin_addr,gClientAddrBuffer);
      feclog("COMMAND %s called by %s from addr %s: access lock refused access",
        Property,cln->userName,gClientAddrBuffer);
      ccerr(access_denied);
    }
    if (putCommandsInFeclog) feclog("COMMAND %s called by %s",Property,cln->userName);
  }
err:
  if (cc == server_redirection && el != NULL)
  {
    memcpy(stsString,el->EqmLastErrorString,STATUS_SIZE);
  }
  *listEntry = el;
  return cc;
}
int renegotiateContract(ContractListStruct *cl,int cc)
{
  CONSUMER *c;
  UINT32 doutsize = 0,fmtoutsize;
  int addext = FALSE;

  if ((fmtoutsize=fmtsizeof(cl->contract.EqmFormatOut)) != 0)
  {
    doutsize = cl->dataSizeOut*fmtoutsize;
    if (cc) addext = TRUE;
    if (LFMT(cl->contract.EqmFormatOut) == CF_SPECTRUM) doutsize += 96;
  }
  for (c=cl->consumer; c != NULL; c=c->next)
  {
    if (addext) doutsize += STSSIZE(c->client->protocol);
    c->status.bytesRemaining = (UINT32)doutsize;
    c->status.bytesTotal = (UINT32)doutsize;
    c->status.nextDataSize = c->status.bytesRemaining;
    if (c->status.bytesRemaining > c->status.mtu) c->status.nextDataSize = c->status.mtu;
    c->status.numblks = (UINT16)(doutsize/c->status.mtu + (doutsize%c->status.mtu ? 1 : 0));
    if (doutsize == 0) c->status.numblks = 1;
  }
  return 0;
}
void gateOutputData(BYTE *data,int size,short format,UINT32 gate,int outputBoolean)
{
  int i;
  UINT16 *sval = (UINT16 *)data;
  UINT32 *ival = (UINT32 *)data;
  if (data == NULL) return;
  switch (format)
  {
    case CF_BYTE:
      for (i=0; i<size; i++) 
      {
        data[i] &= gate;
        if (outputBoolean) data[i] = data[i] == gate ? 1 : 0;
      }
      break;
    case CF_SHORT:
      for (i=0; i<size; i++) 
      {
        sval[i] &= gate;
        if (outputBoolean) sval[i] = sval[i] == gate ? 1 : 0;
      }
      break;
    case CF_LONG:
      for (i=0; i<size; i++) 
      {
        ival[i] &= gate;
        if (outputBoolean) ival[i] = ival[i] == gate ? 1 : 0;
      }
      break;
    default:
      return;
  }
}
int gEqmFcnContractTableId = -1;
TINE_EXPORT int GetLastContractId(void)
{
  if (gEqmFcnContractTableId == -1) return -1;
  return ContractList[gEqmFcnContractTableId]->contractID;
}
TINE_EXPORT int GetLastContractTableId(void)
{
  return gEqmFcnContractTableId;
}
TINE_EXPORT int GetContractAccessRate(int id)
{
  if (id < 0) id = gEqmFcnContractTableId;
  if (id < 0 || id >= ncontracts) return 0;
  return ContractList[id]->PollingRate;
}

int EqmFcnWithWildcardDevice(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access)
{
  DTYPE drdb;
  char sDevName[DEVICE_NAME_SIZE];
  int start = 0, len = 0, isCont = FALSE, isNotReady = FALSE;
  int cc = not_allowed, i, k;
  int fmt, fmtlen, nmlen, noff, voff, soff;
  UINT32 devlistsize = 0;
  ExportPropertyListStruct *prp = GetProperyListStruct(el->EqmName,DevProperty);
  char beginToken[64], endToken[64], midToken[64];
  int blen = 0, elen = 0, mlen = 0;
  NAME64 *devlist=NULL;
  BYTE *rb;
  float *frb;
  UINT32 *lrb;
  NAME64 *nrb = NULL;
  UINT32 nlen;
  double wcTimestamp = 0;

  if (prp == NULL) return illegal_property;
  if (isSpectrumArray(el->EqmName,DevProperty)) return not_allowed;
  if (dout == NULL || dout->dArrayLength < 1) return not_allowed;
  if (!isLocal(el->EqmName,DevName,DevProperty)) return data_not_local;
# ifdef INTEL
  rb = (BYTE *)alloca(dout->dArrayLength*sizeof(float));
# else
  rb = tmpWorkArea;
# endif
  frb = (float *)rb; lrb = (UINT32 *)rb;
  if (prp->devNames != NULL)
  {
	  devlist = prp->devNames;
    devlistsize = prp->numDevices;
  }
# if !defined(SMALL_TINE_LIB)
  else if (el->hasPropertyDeviceNames)
  {
    nlen = prp->prpSize;
#   ifndef VXWORKS
    nrb = (NAME64 *)alloca(2*nlen*sizeof(NAME64)); 
#   else
    if ((len=dout->dArrayLength*sizeof(float)) < tmpWorkAreaSize - 2*nlen*sizeof(NAME64))
      nrb = (NAME64 *)&tmpWorkArea[dout->dArrayLength*sizeof(float)];
#   endif
    if (GetDeviceListFromProperty(el->EqmName,DevProperty,nrb,&nlen) == 0)
    {
      for (i=nlen-1; i>=0; i--) 
      {
        strncpy(nrb[i*2].name,nrb[i].name,DEVICE_NAME_SIZE);
        nrb[i*2].name[DEVICE_NAME_SIZE-1] = 0;
      }
      devlist = (NAME64 *)nrb;
      devlistsize = nlen;
    }
  }
# endif /* not SMALL_TINE_LIB */
  if (devlistsize == 0)
  {
    devlist = el->EqmDeviceList;
    devlistsize = el->EqmNumDevices;
  }
  if ((prp->prpArrayType & AT_CHANNEL) == AT_CHANNEL)
  {
    len = devlistsize;
    isCont = isContiguous(el->EqmName,DevName,devlist,&start,&len);
  }
  switch (dout->dFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      fmtlen = fmtsizeof(dout->dFormat);
      nmlen = fmtlen - 16;
      fmt = prp->prpFormat == CF_FLOAT || prp->prpFormat == CF_DOUBLE ? CF_FLOAT : CF_LONG;
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      if (isCont)
      {
        drdb.data.fptr = frb;
        drdb.dArrayLength = MIN(dout->dArrayLength,(UINT32)len);
        strncpy(sDevName,devlist[start].name,DEVICE_NAME_SIZE);
        cc = (*el->eqmFcn)(sDevName,DevProperty,&drdb,din,access);
        if (cc != 0) return cc;
        wcTimestamp = gDataTimeStamp;
        for (i=0; i<(int)drdb.dArrayLength; i++)
        {
          *(double *)(dout->data.bptr + i*fmtlen + nmlen) = fmt == CF_FLOAT ? (double)frb[i] : (double)lrb[i];
          *(double *)(dout->data.bptr + i*fmtlen + nmlen + 8) = 0;
          strncpy(dout->data.cptr + i*fmtlen,devlist[start+i].name,nmlen);
        }
        dout->dArrayLength = drdb.dArrayLength;
      }
      else
      {
        getWildcardTokens(DevName,beginToken,endToken,midToken);
        /* calculate these once outside the loop ... */
        blen = (int)strlen(beginToken);
        mlen = (int)strlen(midToken);
        elen = (int)strlen(endToken);
        for (i=0, k=0; i<(int)devlistsize && k<(int)dout->dArrayLength; i++)
        {
          if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
          strncpy(sDevName,devlist[i].name,DEVICE_NAME_SIZE);
          drdb.data.fptr = &frb[k];
          drdb.dArrayLength = 1;
          cc = (*el->eqmFcn)(sDevName,DevProperty,&drdb,din,access);
          if (cc == server_redirection) return data_not_local;
          if (cc == not_ready) isNotReady = TRUE;
          if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
          *(double *)(dout->data.bptr + k*fmtlen + nmlen) = fmt == CF_FLOAT ? (double)frb[k] : (double)lrb[k];
          *(double *)(dout->data.bptr + k*fmtlen + nmlen + 8) = (double)cc;
          strncpy(dout->data.cptr + k*fmtlen,devlist[start+i].name,nmlen);
          k++;
        }
        dout->dArrayLength = k;
      }
      gDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    case CF_USTRING:
    case CF_NAME16II:
    case CF_NAME16FI:
      fmt = dout->dFormat == CF_NAME16II ? CF_LONG : CF_FLOAT;
      fmtlen = fmtsizeof(dout->dFormat);
      if (dout->dFormat == CF_USTRING)
      {
        nmlen = 80;
        noff = 16; voff = 4; soff = 0;
      }
      else
      {
        nmlen = fmtlen - 8;
        noff = 0; voff = nmlen; soff = nmlen + 4;
      }
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      if (isCont)
      {
        drdb.data.fptr = frb;
        drdb.dArrayLength = MIN(dout->dArrayLength,(UINT32)len);
        strncpy(sDevName,devlist[start].name,DEVICE_NAME_SIZE);
        cc = (*el->eqmFcn)(sDevName,DevProperty,&drdb,din,access);
        if (cc != 0) return cc;
        wcTimestamp = gDataTimeStamp;
        for (i=0; i<(int)drdb.dArrayLength; i++)
        {
          memset(dout->data.bptr + i*fmtlen,0,fmtlen);
          memcpy(dout->data.bptr + i*fmtlen + voff,&frb[i],4);
          strncpy(dout->data.cptr + i*fmtlen + noff,devlist[start+i].name,16);
        }
        dout->dArrayLength = drdb.dArrayLength;
      }
      else
      {
        getWildcardTokens(DevName,beginToken,endToken,midToken);
        /* calculate these once outside the loop ... */
        blen = (int)strlen(beginToken);
        mlen = (int)strlen(midToken);
        elen = (int)strlen(endToken);
        for (i=0, k=0; i<(int)devlistsize && k<(int)dout->dArrayLength; i++)
        {
          if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
          strncpy(sDevName,devlist[i].name,DEVICE_NAME_SIZE);
          drdb.data.fptr = &frb[k];
          drdb.dArrayLength = 1;
          cc = (*el->eqmFcn)(sDevName,DevProperty,&drdb,din,access);
          if (cc == server_redirection) return data_not_local;
          if (cc == not_ready) isNotReady = TRUE;
          if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
          memset(dout->data.bptr + k*fmtlen,0,fmtlen);
          memcpy(dout->data.bptr + k*fmtlen + voff,&frb[k],4);
          *(UINT32 *)(dout->data.bptr + k*fmtlen + soff) = (UINT32)cc;
          strncpy(dout->data.cptr + k*fmtlen + noff,devlist[i].name,nmlen);
          k++;
        }
        dout->dArrayLength = k;
      }
      gDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    default:
      if (dout->dFormat <= CF_NAME32 && strcmp(DevName,"*") == 0)
      { /* primitive format allowed in this case */
        if (isCont) 
        {
          if (dout->dArrayLength > (UINT32)len) dout->dArrayLength = (UINT32)len;
          strncpy(sDevName,devlist[0].name,DEVICE_NAME_SIZE);
          return (*el->eqmFcn)(sDevName,DevProperty,dout,din,access);
        }
        drdb.dFormat = dout->dFormat;
        drdb.dTag[0] = 0;
        for (i=0; i<(int)devlistsize && i<(int)dout->dArrayLength; i++)
        {
          strncpy(sDevName,devlist[i].name,DEVICE_NAME_SIZE);
          drdb.data.bptr = &dout->data.bptr[i*fmtsizeof(dout->dFormat)];
          drdb.dArrayLength = 1;
          if ((cc=(*el->eqmFcn)(sDevName,DevProperty,&drdb,din,access)) != 0)
          {
            if (cc == not_ready) isNotReady = TRUE; else break;
          }
          if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
        }
        if (cc == server_redirection) return data_not_local;
        dout->dArrayLength = i;
        gDataTimeStamp = wcTimestamp;
        return isNotReady ? not_ready : cc;
      }
  }
  return cc;
}
NAME64 *gPrpList = NULL;
int EqmFcnWithWildcardProperty(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access)
{
  DTYPE drdb;
  char sDevProp[PROPERTY_NAME_SIZE];
  int isNotReady = FALSE;
  int cc = not_allowed, i, k, devnr;
  int fmt, fmtlen, nmlen, noff, voff, soff;
  int prplistsize = 0;
  char beginToken[PROPERTY_NAME_SIZE], endToken[PROPERTY_NAME_SIZE], midToken[PROPERTY_NAME_SIZE];
  int blen = 0, elen = 0, mlen = 0;
  NAME64 *prplist=NULL;
  float frb;
  double wcTimestamp = 0;

  if (dout == NULL || dout->dArrayLength < 1) return not_allowed;
  if ((devnr=GetDeviceNumber(el->EqmName,DevName)) < 0) return illegal_equipment_number;

  if (el->EqmDevicePrpList != NULL && 
      el->EqmDevicePrpList[devnr].prpList != NULL &&
      el->EqmDevicePrpList[devnr].prpListSize > 0)
  {
    prplist = el->EqmDevicePrpList[devnr].prpList;
    prplistsize = (int)el->EqmDevicePrpList[devnr].prpListSize;
  }
# if !defined(SMALL_TINE_LIB)
  else 
  {
    if (gPrpList == NULL)
    {
      if ((gPrpList=(NAME64 *)SystemCalloc(el->EqmNumProperties,sizeof(NAME64))) == NULL) return out_of_server_memory;
      prplistsize = el->EqmNumProperties;
      if ((cc=GetRegisteredPropertyList(el->EqmName,gPrpList,&prplistsize)) != 0) return cc;
    }
    if (dout->dArrayLength > (UINT32)el->EqmNumProperties) dout->dArrayLength = el->EqmNumProperties;
    prplistsize = el->EqmNumProperties;
    prplist = gPrpList;
  }
# endif /* SMALL_TINE_LIB */
  if (prplistsize == 0) return illegal_property;
  switch (dout->dFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      fmtlen = fmtsizeof(dout->dFormat);
      nmlen = fmtlen - 16;
      fmt = CF_FLOAT;
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      getWildcardTokens(DevProperty,beginToken,endToken,midToken);
      /* calculate these once outside the loop ... */
      blen = (int)strlen(beginToken);
      mlen = (int)strlen(midToken);
      elen = (int)strlen(endToken);
      for (i=0, k=0; i<prplistsize && k<(int)dout->dArrayLength; i++)
      {
        if (!wildcardMatches(prplist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
        strncpy(sDevProp,prplist[i].name,32);
        if (isSpectrumArray(el->EqmName,sDevProp)) continue;
        drdb.data.fptr = &frb;
        drdb.dArrayLength = 1;
        cc = (*el->eqmFcn)(DevName,sDevProp,&drdb,din,access);
        if (cc == illegal_format || cc == illegal_equipment_number || cc == dimension_error) continue;
        if (cc == not_ready) isNotReady = TRUE;
        if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
        *(double *)(dout->data.bptr + k*fmtlen + nmlen) = fmt == CF_FLOAT ? (double)frb : (double)(*(UINT32 *)&frb);
        *(double *)(dout->data.bptr + k*fmtlen + nmlen + 8) = (double)cc;
        strncpy(dout->data.cptr + k*fmtlen,prplist[i].name,nmlen);
        k++;
      }
      dout->dArrayLength = k;
      gDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    case CF_USTRING:
    case CF_NAME16II:
    case CF_NAME16FI:
      fmtlen = fmtsizeof(dout->dFormat);
      fmt = dout->dFormat == CF_NAME16II ? CF_LONG : CF_FLOAT;
      if (dout->dFormat == CF_USTRING)
      {
        nmlen = 80;
        noff = 16; voff = 4; soff = 0;
      }
      else
      {
        nmlen = fmtlen - 8;
        noff = 0; voff = nmlen; soff = nmlen + 4;
      }
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      getWildcardTokens(DevProperty,beginToken,endToken,midToken);
      /* calculate these once outside the loop ... */
      blen = (int)strlen(beginToken);
      mlen = (int)strlen(midToken);
      elen = (int)strlen(endToken);
      for (i=0, k=0; i<prplistsize && k<(int)dout->dArrayLength; i++)
      {
        if (!wildcardMatches(prplist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
        strncpy(sDevProp,prplist[i].name,32);
        if (isSpectrumArray(el->EqmName,sDevProp)) continue;
        drdb.data.fptr = &frb;
        drdb.dArrayLength = 1;
        cc = (*el->eqmFcn)(DevName,sDevProp,&drdb,din,access);
        if (cc == illegal_format || cc == illegal_equipment_number || cc == dimension_error) continue;
        if (cc == not_ready) isNotReady = TRUE;
        if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
        memset(dout->data.bptr + k*fmtlen,0,fmtlen);
        memcpy(dout->data.bptr + k*fmtlen + voff,&frb,4);
        *(UINT32 *)(dout->data.bptr + k*fmtlen + soff) = (UINT32)cc;
        strncpy(dout->data.cptr + k*fmtlen + noff,prplist[i].name,nmlen);
        k++;
      }
      dout->dArrayLength = k;
      gDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    default:
      break;
  }
  return cc;
}
int callStk(ContractListStruct *cl,char *dataOut,char *dataIn,int (*fcn)(CONTRACT *,char *,char *),int runInStkThread)
{
  int cc;
  CONTRACT *contract = &cl->contract;
# ifdef MULTITHREADED
  static MXHANDLE hMx = 0;
# endif

  cl->isStkCal = TRUE;
  if (gStockRecourceSet.isInfoLocked) return not_ready;
# ifdef MULTITHREADED
  if (useMultiThreadedStockFunctions && runInStkThread)
  {
    if (cl->compStatus == not_signalled && hMx == 0) 
    {
      if (tineDebug > 2) thread_dbglog("release stock mutex");
      gStockRecourceSet.isInfoLocked = TRUE;
      cl->callPending = TRUE;
      gStockRecourceSet.StkCrs.dout.dArrayLength = contract->EqmSizeOut;
      gStockRecourceSet.StkCrs.dout.dFormat = LFMT(contract->EqmFormatOut);
      strncpy(gStockRecourceSet.StkCrs.dout.dTag,contract->strTagOut,TAG_NAME_SIZE);
      gStockRecourceSet.StkCrs.dout.data.cptr = dataOut;
      gStockRecourceSet.StkCrs.din.dArrayLength = contract->EqmSizeIn;
      gStockRecourceSet.StkCrs.din.dFormat = LFMT(contract->EqmFormatIn);
      strncpy(gStockRecourceSet.StkCrs.din.dTag,contract->strTagIn,TAG_NAME_SIZE);
      gStockRecourceSet.StkCrs.din.data.cptr = dataIn;
      gStockRecourceSet.StkCls = cl;
      gStockRecourceSet.fcn = fcn;
      hMx = gStockRecourceSet.hStkCallMutexPair[gStockRecourceSet.StkCallMutexIndex];
      cc = ReleaseMutex(hMx);
      millisleep(0);
    }
    if (!cl->callPending) return not_ready; /* only one stock call at a time */
    cc = WaitForMutex(hMx,gSystemTick);
    if (cc != 0)
    {
      if (tineDebug > 2) thread_dbglog("timed out waiting for stock mutex");
      cc = not_ready; 
    } 
    else
    {
      if (tineDebug > 2) thread_dbglog("got stock mutex");
      cc = cl->compStatus;
      cl->callPending = FALSE;
      hMx = 0;
    }
  }
  else
# endif
  {
    gStockRecourceSet.isInfoLocked = TRUE;
    gStockRecourceSet.StkCrs.dout.dArrayLength = contract->EqmSizeOut;
    gStockRecourceSet.StkCrs.dout.dFormat = LFMT(contract->EqmFormatOut);
    strncpy(gStockRecourceSet.StkCrs.dout.dTag,contract->strTagOut,TAG_NAME_SIZE);
    gStockRecourceSet.StkCrs.dout.data.cptr = dataOut;
    gStockRecourceSet.StkCrs.din.dArrayLength = contract->EqmSizeIn;
    gStockRecourceSet.StkCrs.din.dFormat = LFMT(contract->EqmFormatIn);
    strncpy(gStockRecourceSet.StkCrs.din.dTag,contract->strTagIn,TAG_NAME_SIZE);
    gStockRecourceSet.StkCrs.din.data.cptr = dataIn;
    gStockRecourceSet.StkCls = cl;
    cc = (*fcn)(contract,dataIn,dataOut);
    gStockRecourceSet.isInfoLocked = FALSE;
  }
  return cc;
}

int callEqm(ContractListStruct *cl,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int cc;
  ExportListStruct *el;
# ifdef MULTITHREADED
  static MXHANDLE hMx;
# endif

  if (cl == NULL) return code_failure;
  if ((el=cl->exp) == NULL) return non_existent;
# ifdef MULTITHREADED
  if (useMultiThreadedEquipmentFunctions && callEqmInSeparateThread(el->EqmName,devProperty))
  {
    if (gSystemUseDataProtection) WaitForMutex(el->hDataProtectionMutex,-1);
    if (cl->compStatus == not_signalled) 
    {
      if (tineDebug > 2) thread_dbglog("release mutex");
      gEqmFcnContractTableId = cl->listId;
      el->EqmCls = (void *)cl;
      strncpy(el->EqmCrs.devName,devName,DEVICE_NAME_MAXSIZE);
      strncpy(el->EqmCrs.devProp,devProperty,PROPERTY_NAME_SIZE);
      memcpy(&el->EqmCrs.dout,dout,sizeof(DTYPE));
      memcpy(&el->EqmCrs.din,din,sizeof(DTYPE));
      el->EqmCrs.access = access;
      el->eqmbusy = 1;
      hMx = el->hEqmCallMutexPair[el->EqmCallMutexIndex];
      ReleaseMutex(hMx);
    }
    cc = WaitForMutex(hMx,gSystemTick);
    if (cc != 0)
    {
      if (tineDebug > 2) thread_dbglog("timed out waiting for mutex");
      el->eqmbusy++;
      cc = not_ready; 
    } 
    else
    {
      if (tineDebug > 2) thread_dbglog("got mutex");
      el->eqmbusy = 0;
      /* did the call actually return 'not_signalled' ? */
      cc = cl->compStatus == not_signalled ? not_ready : cl->compStatus;
    }
    if (gSystemUseDataProtection) ReleaseMutex(el->hDataProtectionMutex);
  }
  else
# endif
  {
    el->eqmbusy = 1;
    el->EqmCls = (void *)cl;
    gEqmFcnContractTableId = cl->listId;
    if (isWildcardContract(devName))
    {
      cc = EqmFcnWithWildcardDevice(el,devName,devProperty,dout,din,access);
    }
    else if (isWildcardContract(devProperty))
    {
      cc = EqmFcnWithWildcardProperty(el,devName,devProperty,dout,din,access);
    }
    else
    {
      cc = (*el->eqmFcn)(devName,devProperty,dout,din,access);
    }
    el->eqmbusy = 0;
  }
  return cc;
}

int doContract(ContractListStruct *cl)
{
  CONTRACT *con = &cl->contract;
  ExportListStruct *el = cl->exp;
  BYTE *dataIn = cl->DataIn;
  BYTE *dataOut = cl->DataBuf;
  int i,fmtsiz,cc=0,rw=1,isGated=FALSE;
  short access;
  UINT32 doutlen,doutsize,dp,gate=0;
  struct structStruct *s;
  char txtbuf[128],tDevProperty[PROPERTY_NAME_SIZE+4];
  double dts = gServerDataTimeStamp; /* gServerDataTimeStamp is set prior to call */
  static DTYPE din,dout;
  static char sDevName[DEVICE_NAME_MAXSIZE],sDevProperty[PROPERTY_NAME_SIZE+4];  /* static due to multithreaded eqms */

  doutlen = con->EqmSizeOut; fmtsiz = fmtsizeof(con->EqmFormatOut);
  doutsize = doutlen*fmtsiz; if (LFMT(con->EqmFormatOut) == CF_SPECTRUM) doutsize += 96;

  if (tineDebug > 3)
  {
    strncpy(txtbuf,con->EqmName,EQM_NAME_SHORTSIZE); txtbuf[EQM_NAME_SHORTSIZE] = 0; strcat(txtbuf," ");
    strncat(txtbuf,con->EqmProperty,PROPERTY_NAME_SIZE); strcat(txtbuf," ");
    strncat(txtbuf,con->EqmDeviceName,DEVICE_NAME_SIZE);
    if (con->EqmSizeIn != 0)
      DataDump(txtbuf,(char *)dataIn,(int)con->EqmSizeIn,(short)LFMT(con->EqmFormatIn),(short)RTOL);
  }

  if (cl->expired)
  {
    if (tineDebug > 1) dbglog("%s %s %s -> returning expired data",
      cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty);
    con->EqmSizeOut = cl->dataSizeOut;
    ccerr(cl->returnCode);
  }

  /* check stock properties: */
  /* copy contract's property into temporary variable */
  if (strchr(con->EqmProperty,'*') != NULL && isSimpleFormat((short)LFMT(con->EqmFormatOut))) strcpy(tDevProperty,"PROPERTIES");
  else strncpy(tDevProperty,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((i=GetStockPropertyId(tDevProperty)) >= 0)
  { /* zero out the output buffer if call is fresh */
    if (doutsize && !cl->callPending) memset(dataOut,0,(size_t)doutsize);
    cc = callStk(cl,(char *)dataOut,(char *)dataIn,StockProperty[i].fcn,StockProperty[i].runInSeparateThread);
    ccerr(cc);
  }
  /* copy contract's property into temporary variable */
  strncpy(tDevProperty,con->EqmProperty,PROPERTY_NAME_SIZE);
  /* check stock modifiers: (allow overloads) */
  if (GetPropertyId(con->EqmName,con->EqmProperty) < 0)
  { /* not a registered property */
    if (el->EqmPropertyQueryFcn == NULL && el->XEqmPropertyQueryFcn == NULL)
    { /* there's no registered property query function */
      for (i=0; i<NrStockModifiers; i++)
      { /* run through the list of stock modifiers */
        if (strstr(con->EqmProperty,StockModifier[i].modName))
        { /* found a match */
          if (doutsize && !cl->callPending) memset(dataOut,0,(size_t)doutsize);
          if ((isGated=IsGatedModifier(con,tDevProperty,&gate)) != 0) break;
          cc = callStk(cl,(char *)dataOut,(char *)dataIn,StockModifier[i].fcn,StockModifier[i].runInSeparateThread);
          ccerr(cc);
        }
      }
    }
  }

  if (el == NULL || el->eqmFcn == NULL) return non_existent;

  if (cl->compStatus == not_signalled && el->eqmbusy) 
  {
    if (tineDebug > 2) dbglog("%s %s %s is %s (busy count %d)",
      cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,
      cl->compStatus == not_signalled ? "not signalled" : "busy", el->eqmbusy);
    return el->eqmbusy > 1 ? cl->compStatus = operation_busy : not_ready;
  }

  /* copy contract's property into local static variable */
  strncpy(sDevProperty,tDevProperty,PROPERTY_NAME_SIZE);
  /* copy contract's device name into local static variable */
  strncpy(sDevName,con->EqmDeviceName,DEVICE_NAME_SIZE); sDevName[DEVICE_NAME_SIZE] = 0;
  if (strlen(sDevName) == DEVICE_NAME_SIZE && sDevName[DEVICE_NAME_SIZE-1] == '&') 
  { /* long device name requested */
    sDevName[DEVICE_NAME_SIZE-1] = 0;
    if (cl->extStringSpace != NULL) 
    {
      strncat(sDevName,cl->extStringSpace,cl->extStringSize);
    }
    else
    {
      if (tineDebug) dbglog("non-terminated long device name !");
    }
    sDevName[DEVICE_NAME_MAXSIZE-1] = 0;
  }

  /* call equipment function: */
  if (el->eqmFcn != NULL)      /* EQM function registered */
  {
    dout.dStamp = dout.sysStamp = 0;
    dout.dTimeStamp = dts;
    dout.dArrayLength = con->EqmSizeOut;
    dout.dFormat = LFMT(con->EqmFormatOut);
    dout.data.vptr = dataOut;
    strncpy(dout.dTag,con->strTagOut,TAG_NAME_SIZE);
    if (dout.dFormat == CF_STRUCT)
    {
      AssignStructDataPtr(dout.data.vptr,dataOut,NONE);
      dout.dArrayLength = i;
    }
    if (SystemPresetMemory && doutsize && cl->compStatus == not_signalled) memset(dataOut,0,(size_t)doutsize);
    din.dStamp = din.sysStamp = 0;
    din.dArrayLength = con->EqmSizeIn;
    din.dFormat = LFMT(con->EqmFormatIn);
    din.data.vptr = dataIn;
    strncpy(din.dTag,con->strTagIn,TAG_NAME_SIZE);
    if (din.dFormat == CF_STRUCT)
    {
      AssignStructDataPtr(din.data.vptr,dataIn,RTOL);
      din.dArrayLength = i;
    }
    access = con->EqmAccess;
    if (access & CA_WRITE && currentAccessLock.lockType) access |= CA_LOCKED;
    cc = callEqm(cl,sDevName,sDevProperty,&dout,&din,access);
    
    if (cc != 0 && !(cc & CE_SENDDATA)) ccerr(cc);
    if (dout.dTimeStamp != dts) gServerDataTimeStamp = dout.dTimeStamp;
    cl->datastamp = dout.dStamp;
    if (dout.dFormat == CF_STRUCT)
    {
      AssignStructDataPtr(dataOut,dataOut,LTOR);
    }
    if (dout.dArrayLength < con->EqmSizeOut)
    {
      if (!SystemPresetMemory) /* zero out the remainder for later data comparisons */
      {
        dp = dout.dArrayLength*fmtsiz;
        memset(&dataOut[(int)dp],0,(size_t)(con->EqmSizeOut-dout.dArrayLength)*fmtsiz);
      }
      con->EqmSizeOut = dout.dArrayLength;
    }
  }
err:
  if (isGated) gateOutputData(dataOut,(int)con->EqmSizeOut,(short)LFMT(con->EqmFormatOut),gate,isGated == GATE_BOOLEAN);
  if (con->EqmAccess&CA_FIRST) con->EqmAccess &= ~(CA_FIRST);
  if (cc == not_ready) return cc; /* do NOT pass go, do NOT collect $200 */
  if (cc < 0) cc = ambiguous;
  if (tineDebug > 3 && (cc == 0 || cc & CE_SENDDATA))
    DataDump(txtbuf,(char *)dataOut,(int)con->EqmSizeOut,(short)LFMT(con->EqmFormatOut),(short)LTOR);

  cl->dataSizeOut = con->EqmSizeOut;    /* record possible new data output length */
  if (doutlen != con->EqmSizeOut)
  {
    con->EqmSizeOut = doutlen;   /* don't disrupt the contract itself */
    cc |= CE_RENEGOTIATE;
  }
  return cc;
}
int initSendNetGlobalsList(void)
{
  int cc=0,i,n=1,fmtsize,dsize;
  int GlobalPacketSize = 0;
  static int initialized = 0;

  if (initialized) return 0; else initialized = 1;
  GlobalPacketSize = 0;
  if (StartupDebug) dbglog("sending the following global parameters :");
  for (i = 0, GlobalPacketSize = sizeof(HeaderStruct);
       i < ngcasts && GlobalPacketSize < gBrdcstBufferSize; i++)
  {
    if (StartupDebug) dbglog("keyword : %s -> %d %s value(s)",gcastTbl[i]->keyword,gcastTbl[i]->size,itof(gcastTbl[i]->format));
    GlobalPacketSize += sizeof(GlobalDataHdr);
    fmtsize = fmtsizeof(gcastTbl[i]->format);
    if (!fmtsize) ccerr(illegal_format);
    if ((dsize=gcastTbl[i]->size*fmtsize) >
         (int)(gBrdcstBufferSize - sizeof(HeaderStruct) - sizeof(GlobalDataHdr)))
    {
      if (StartupDebug) dbglog("globals %s packet exceeds %d bytes\n>",
	      gcastTbl[i]->keyword,gBrdcstBufferSize);
      ccerr(buffer_too_small);
    }
    GlobalPacketSize += dsize;
    if (GlobalPacketSize > gBrdcstBufferSize)
    {
      n++;
      GlobalPacketSize = sizeof(HeaderStruct);
    }
  }
  if (StartupDebug) dbglog("total globals packets required : %d",n);
err:
  if (cc)
  {
    if (StartupDebug) dbglog("initSendNetGlobalsList : %s",erlst[cc]);
    ngcasts = 0;
  }
  return 0;
}
int initServerTransport(void)
{
  int cc = 0;
  char txt[128];
  if ((cc=initControlStructs()) != 0) 
  {
    sprintf(txt,"Cannot map header structures");
    ccerr(cc);
  }
# if defined(IPXBSD) || defined(IPXNW)
  if ((cc=initIPX()) != 0)
  {
    sprintf(txt,"IPX/SPX server failed to initialize");
    ccerr(cc);
  }
# endif
  if ((cc=initTCP()) != 0)
  {
    sprintf(txt,"UDP/TCP server failed to initialize");
    ccerr(cc);
  }
err:
  if (cc)
  {
    feclog("%s : %s",txt,erlst[cc]);
  }
  return cc;
}
int initTineServices(void)
{
  int cc = 0;
  if (gInitTineServicesDone) return 0;
  if (!FecNameRegistered)
  {
    feclog("aborting server initialization : fec name is unknown");
    return host_not_resolved;
  }
  getSrvIPAddr();
  if (gRespondToServiceRequests) cc = RegisterLocalServiceModule();
  cc = initServerTransport();
  initAlarmServer();
  makeStockPropXRefTable();
  memset(&currentAccessLock,0,sizeof(currentAccessLock));
  SysPoll = MaxPollingRate;
  if (cdiSystemInitHook != 0) cdiSystemInitHook(); /* if CDI is in use ... */
# if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
  if (CreateShmWindow()) feclog("Cannot create shm window!");
# endif
  initClient();
  gInitTineServicesDone = TRUE;
  return cc;
}
int initTineResources(void)
{
  int cc=0;
  BYTE *ver = SystemVersion();

  feclog("VERSION : %d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  feclog("OS      : %s",OS_TAG);
  if (srvWorkAreaSize >= WORK_SIZE && 
      MaxSystemTransportSize > MAX_TRANSPORTSIZE &&
      MaxSystemTransportSize > srvWorkAreaSize) srvWorkAreaSize = MaxSystemTransportSize;
  tmpWorkAreaSize = srvWorkAreaSize;
# ifdef MULTITHREADED
  feclog("Work Area: %u",srvWorkAreaSize);
  feclog("Temp Size: %u",tmpWorkAreaSize);
  feclog("FEC: is running multi-threaded");
# else
  if (srvWorkAreaSize <= WORK_SIZE && 
      MaxSystemTransportSize > srvWorkAreaSize) MaxSystemTransportSize = srvWorkAreaSize;
  feclog("Work Area: %u",(unsigned int)srvWorkAreaSize);
  feclog("Temp Size: %u",(unsigned int)tmpWorkAreaSize);
  feclog("Max Transport Size: %u",(unsigned int)MaxSystemTransportSize);
# ifdef INTEL
  tmpWorkArea = srvWorkArea;
# endif
# endif
  feclog("FEC HOME : [%s]",FecDBpath);
  startuptime = time(NULL);
  nconsumers = ncontracts = 0;
  memset(gFecDBSubDir,0,20);
  if ((srvWorkArea = (BYTE *)SystemCalloc(1,(size_t)srvWorkAreaSize)) == NULL)
  {
   feclog("Work Space %u bytes: insufficient memory",(unsigned int)srvWorkAreaSize);
   return out_of_server_memory;
  }
  if (tmpWorkArea == NULL && (tmpWorkArea=(BYTE *)SystemCalloc(1,tmpWorkAreaSize)) == NULL)
  {
    feclog("Temp Work Space: insufficient memory");
    return out_of_server_memory;
  }
  if (MaxNumClients <= 0) return 0; /* no server functionality required */
  if ((ContractList=(ContractListStruct **)SystemCalloc(MaxNumContracts,sizeof(ContractListStruct *))) == NULL)
  {
   feclog("Contract List: insufficient memory");
   return out_of_server_memory;
  }
  if ((Consumers=(CLN **)SystemCalloc(MaxNumClients,sizeof(CLN *))) == NULL)
  {
   feclog("Client List: insufficient memory");
   return out_of_server_memory;
  }
  if (canSendGlobals)
  {
    if ((gcastTbl=(GlobalListStruct **)SystemCalloc(MaxNumGlobals,sizeof(GlobalListStruct *))) == NULL)
    {
      feclog("Globals List: insufficient memory");
      return out_of_server_memory;
    }
  }
  GetAliasTable();
  memset(&IPXFecAddress,0,sizeof(IPXAddress));
  getRegisteredExports(EXPORT_FILE,NULL);
  getFecName();
  getRegisteredUsers(&gAllowedUsersList,USERS_FILE,&NrOfUsers,"WRITE",&WRITEaccess);
  getRegisteredUsers(&gDeniedUsersList,DENYUSER_FILE,&NrOfDeniedUsers,"DENIED",&DENYaccess);
  if (!FecNameRegistered)
  {
    gDelayInitTineServices = TRUE;
    feclog("delaying system initialization");
  }
  else
  {
    cc = initTineServices();
  }
  return cc;
}
TINE_EXPORT int ServerExitCondition = 0;
TINE_EXPORT void CloseServers(int closesockets)
{
  closesockets = TRUE; /* always true */
# if defined(IPXNW) || defined(IPXBSD)
  closeIPXServer(); 
# endif
  closeIPServer();
# if defined(WIN32)
  stopPipeListener();
# endif
  gInitTineServicesDone = FALSE;
  gInitIPServicesDone = FALSE;
  FecNameRegistered = FALSE;
  gInitIPClientDone = FALSE;
  SystemRunning = FALSE;
  return;
}
void quitServer(void)                         /* allow a clean exit */
{
  ServerExitCondition = 1;
  return;
}
char gLastWriteAccessUsr[USERNAME_SIZE+4] = "";
char gLastWriteAccessAddr[ADDR_SIZE+4] = "";
char gLastWriteAccessPrp[PROPERTY_NAME_SIZE+4] = "";
char gLastWriteAccessDev[DEVICE_NAME_SIZE+4] = "";
double gLastWriteAccessTime = 0;
UINT32 gBurstLimitReachedCount = 0;
UINT32 gTimeServerCycle = 0;
UINT32 gTimeServerIdleLastCycle = 0;
int gLastMinuteInitializationDone = FALSE;
TINE_EXPORT void ResetLastMinuteInitialization(void) 
{ 
  gLastMinuteInitializationDone = FALSE; 
  memset(gFecDBSubDir,0,20);
}
TINE_EXPORT void resetLastMinuteInitialization(void) { ResetLastMinuteInitialization(); }
int gCycleCounts = 0, gMaxCycleCounts = 1;
int serverCycle(void)
{
  static struct timeval lastCycleTime;
  static int CycleCounts = 0;

  if (MaxNumClients <= 0) return 0; /* no server functionality required */
  if (ServerExitCondition) { if (ServerExitCondition++ > 3) haltServer(); }
  if (!ServerInitialized) return 0; /* no server functionality required yet */
  if (!gLastMinuteInitializationDone) 
  {
    gettimeofday(&thisCycleTime,(struct timezone *)NULL);/* Cycle time */
    gLastMinuteInitializationDone = LastMinuteInitialization();
    lastCycleTime = thisCycleTime;
    if (useGlobalSynchronization) SystemStartGlobalSynchronization();
	SystemRunning = TRUE;
  }
# ifdef MULTITHREADED
  if (gNeedToStartEqmThreads) StartEqmThreads();
# endif
  if (WaitForSystemMutex(hSystemServerMutex,0) != 0) return not_ready;
  gettimeofday(&thisCycleTime,(struct timezone *)NULL);/* Cycle time */
  gTimeServerCycle += MSECS(thisCycleTime,lastCycleTime);
  lastCycleTime = thisCycleTime;
  AcceptRequest(&thisCycleTime);/* Accept Consumer subscriptions and renewals   */
  gTimeServerIdleLastCycle += MSECS(thisCycleTime,lastCycleTime);
  doScheduler(&thisCycleTime);  /* Is it time to service consumers?   */
  if (ngcasts) doGCast(&thisCycleTime);
  if (acquireDeviceNamesFromProperties) devNamesInitialization();
  Cleanup(&thisCycleTime);
  SurveyAlarms(&thisCycleTime);
  if (!ExportsRegistered) sndRegisteredExports(&thisCycleTime);
  if (currentAccessLock.lockType) checkAccessLocks((time_t)thisCycleTime.tv_sec);
  if (gRespondToServiceRequests) reAttachToServiceGroup();
  gTimeServerCycle += MSECS(thisCycleTime,lastCycleTime);
  lastCycleTime = thisCycleTime;
  CycleCounts++;
  if (gTimeServerCycle >= 1000)
  {
    if (CycleCounts > gMaxCycleCounts) gMaxCycleCounts = CycleCounts;
    gCycleCounts = CycleCounts;
    if (ServerCycleMode != POLLING)
    {
      AveBusyTime = (int)((gTimeServerCycle-gTimeServerIdleLastCycle)*100/gTimeServerCycle);
    }
    else
    {
      AveBusyTime = (gMaxCycleCounts-CycleCounts)*100/gMaxCycleCounts;
    }
    gTimeServerCycle = gTimeServerIdleLastCycle = 0;
    CycleCounts = 0;
  }
  ReleaseSystemMutex(hSystemServerMutex);
  return 0;
}

ContractListStruct **ContractList;
CLN **Consumers;                /* registered consumer list */
int nconsumers = 0, ncontracts = 0;

int StaleData = 0;

char *SubType[] = {"CANCEL","SINGLE","REFRESH","POLL"};
int RemoveConsumerFromList(CONSUMER *c,int j)          /* from contract j */
{
  CONSUMER *p=ContractList[j]->consumer;
  int wasPending = FALSE;

  if (j < 0 || j>=ncontracts || c == NULL) return -2;

  if (tineDebug > 1) dbglog("remove consumer %s from contract %d",c->client->userName,j);
  if (c->status.blknum < c->status.numblks) wasPending = TRUE;
  c->client->count = MAX(c->client->count-1,0);
  StaleData++; /*  Force RemoveConsumer() */
  if (p == c) ContractList[j]->consumer = c->next;
  else
  {
    while (p->next != c) p=p->next;
    p->next = c->next;
  }
  if (p == p->next) p->next = NULL; /* just in case ! */
  SystemFree((char *)c);
  ContractList[j]->nconsumers = MAX(ContractList[j]->nconsumers-1,0);
  if (ContractList[j]->nconsumers == 0) return 0;
  if (wasPending) ContractList[j]->deliveryPending--;
  p = ContractList[j]->consumer;
  ContractList[j]->PollingRate = p->status.PollingRate;
  for (p=p->next; p != NULL; p=p->next)
   ContractList[j]->PollingRate =
     findpoll(ContractList[j]->PollingRate,p->status.PollingRate);
  return 0;
}
int RemoveConsumer(int i)  /* return memory to heap */
{
  if (i < 0 || i >= nconsumers) return -2;
  if (Consumers[i] == NULL) return -3;
  if (tineDebug > 1) dbglog("remove consumer %s",Consumers[i]->userName);
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  if (Consumers[i]->transport == PIPE && Consumers[i]->sck > 0) 
  {
    close(Consumers[i]->sck);
    if (nipcconsumers > 0) nipcconsumers--;
  }
# endif  
  SystemFree((char *)Consumers[i]);
  for (;i<nconsumers-1; i++) Consumers[i] = Consumers[i+1];
  Consumers[--nconsumers] = NULL;
  return 0;
}
int RemoveContract(int j)  /* return memory to heap */
{
  CONSUMER *c=ContractList[j]->consumer,*p;

  if (j < 0 || j>=ncontracts) return argument_list_error;
  if (ContractList[j]->compStatus == not_posted) 
  {
    if (tineDebug) dbglog("remove contract %d refused (not yet posted)",j);
    return ContractList[j]->compStatus;
  }
  if (tineDebug > 1) dbglog("remove contract %d",j);
  for (p=c; p != NULL; c=p)                
  { /* decrement consumer count: */
    c->client->count = MAX(c->client->count-1,0);
    StaleData++; /*  Force RemoveConsumer() */
    p = c->next;
    SystemFree((char *)c);
  }
  if (ContractList[j]->expired == CONTRACT_ACTIVE) nExpiredContracts++;
  ContractList[j]->expired = CONTRACT_AVAILABLE;
  if (ContractList[j]->DataIn) 
  {
    SystemFree((char *)ContractList[j]->DataIn);
    ContractList[j]->DataIn = NULL;
  }
  if (ContractList[j]->DataOut) 
  {
    SystemFree((char *)ContractList[j]->DataOut);
    ContractList[j]->DataOut = NULL;
  }
  if (ContractList[j]->extStringSpace) 
  {
    SystemFree((char *)ContractList[j]->extStringSpace);
    ContractList[j]->extStringSpace = NULL;
  }
  for (j=0, SysPoll = MaxPollingRate; j<ncontracts; j++)
  {
    if (ContractList[j]->expired) continue;
    SysPoll = findpoll(SysPoll,ContractList[j]->PollingRate);
  }
  return 0;
}
UINT32 gContractMisses = 0;
UINT32 gContractDelays = 0;
UINT32 gClientMisses = 0;
UINT32 gClientReconnects = 0;
UINT32 gClientRetries = 0;
UINT32 gSystemTick = SYSTEM_TICK;
UINT32 gSynchronizeContracts = (UINT32)CONTRACT_SYNCHRONIZE;
int doScheduler(struct timeval *tv)
{
  ContractListStruct *cl; /* improve readability by assignment */
  CONSUMER *c;
  short cc,rc,i,dirty=0,Stale,fmtoutsize;
  UINT32 doutsize;
  SINT32 delta_t;
  double dts;
  int isScheduled, isbusy, msec;
  struct timeval tv_stamp;                

  if (WaitForSystemMutex(hSystemKernelMutex,gSystemTick) != 0) return not_ready;
  for (i=0; i<ncontracts; i++)
  {
    if ((cl=ContractList[i])->expired == CONTRACT_AVAILABLE) continue;/* just a place-holder */
    if (cl->nconsumers == 0 && cl->expired == CONTRACT_ACTIVE) 
    { /* no clients, not marked as expired */
      if (cl->contract.EqmAccess & CA_LAST) 
      { /* going out of scope */
        if (tineDebug > 1)
        {
          dbglog("[%d] (%s) %s %s going out of scope with no clients",
            i,cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty);
        }
        doContract(cl); 
      }
      RemoveContract(i);
      continue;
    }
    cc = Stale = isScheduled = 0;
    if ((fmtoutsize=fmtsizeof(cl->contract.EqmFormatOut)) != 0)
    {
      doutsize = cl->contract.EqmSizeOut*fmtoutsize;
      if (LFMT(cl->contract.EqmFormatOut) == CF_SPECTRUM) doutsize += 96;
    }
    else doutsize = 0;
    if ((delta_t=MSECS(*tv,cl->lasttime)) < 0) cl->lasttime = *tv; /* system clock set back ! */
    if (delta_t <= cl->PollingRate - (SINT32)gSystemTick) continue;
    if (cl->expired == CONTRACT_EXPIRED && cl->nconsumers == 0)
    {
      if (delta_t > 3*cl->PollingRate) RemoveContract(i);
      continue;
    }
    if (delta_t < MAXALLOWED_POLLRATE && delta_t >= 2*cl->PollingRate)
    {
      if (tineDebug > 1)
      {
        dbglog("[%d] (%s) %s %s missed last schedule (%d vs %d)",
          i,cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,
          (int)delta_t,(int)(2*cl->PollingRate));
      }
      cl->misses++;
      gContractMisses++;
    }
    if (cl->deliveryPending > 0) 
    {
      if (tineDebug > 1)
      {
        dbglog("[%d] (%s) %s %s still pending",
          i,cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty);
      }
      cl->delays++;
      gContractDelays++;
      continue; /* still sending out the last round */
    }
    if (cl->starttime.tv_sec == 0 && cl->starttime.tv_usec == 0) cl->starttime = *tv;
    tv_stamp = *tv; 
    msec = (int)(tv->tv_usec/1000);
    /* Synchronize on 1000 msec boundaries if possible */
    if (!(cl->PollingRate%1000) && 
        gSynchronizeContracts   && 
        msec < (1000 - (int)gSystemTick)) tv_stamp.tv_usec = 0;
    dts = PutDataTimeStamp(gDataTimeStampOffset,tv->tv_sec,msec);
    setDataTimeStamp(dts);
    if (cl->nconsumers == 1 && cl->consumer->status.counter == 1)
      cl->contract.EqmAccess |= CA_LAST; /* dangling client going out of scope */
    switch (cc=doContract(cl))
    {
      case 0: break;
      case not_ready:
        msec = MSECS(*tv,cl->starttime);
        isbusy = cl->isStkCal ? gStockRecourceSet.isThreadBusy : cl->exp->eqmbusy;
        if (!isbusy || (msec < cl->PollingRate + 500 && isbusy))
        { /* call returned not_ready or still not_signalled */
          for (c=cl->consumer; c != NULL; c=c->next) c->status.Stale = 0;
          cl->deliveryPending = 0;
          continue; /* don't send out */
        }
        if (tineDebug > 2) dbglog("contract [%d] busy for %d msecs (busy count %d)",i,msec,cl->exp->eqmbusy);
        goto err;
      case call_redirection:
      case server_redirection:
        memcpy(cl->compString,cl->exp->EqmLastErrorString,STATUS_SIZE);
      default: 
        if (cc & 0xf000) 
        { /* CE_RENEGOTIATE or CE_SENDDATA */
          if ((rc=renegotiateContract(cl,(cc & 0xfff))) != 0) cc = rc; else cc &= ~CE_RENEGOTIATE;
        }
        else goto err;
    }
    if (cl->contract.EqmAccess & CA_WRITE)
    {
      if (cl->consumer != NULL && cl->consumer->client != NULL) 
      {
        strncpy(gLastWriteAccessUsr,cl->consumer->client->userName,USERNAME_SIZE);
        strncpy(gLastWriteAccessAddr,inet_ntoa(cl->consumer->client->IPaddress.sin_addr),ADDR_SIZE);
      }
      strncpy(gLastWriteAccessPrp,cl->contract.EqmProperty,PROPERTY_NAME_SIZE);
      strncpy(gLastWriteAccessDev,cl->contract.EqmDeviceName,DEVICE_NAME_SIZE);
      gLastWriteAccessTime = gDataTimeStamp;
    }
    if (doutsize && cl->expired == CONTRACT_ACTIVE && memcmp(cl->DataOut,cl->DataBuf,(size_t)doutsize))
    { /* Data have changed: the cached data are dirty */
      memcpy(cl->DataOut,cl->DataBuf,(size_t)doutsize);
      Stale = dirty = 1;                       
    }                                          
    else
    { /* Cached data have not changed */
      if (MSECS(*tv,cl->lastmarked) > (HEARTBEAT-1)*1000L) Stale = 1;
      dirty = 0;                               
    }

err:if (cc) Stale = 1;
    if (Stale) cl->lastmarked = tv_stamp; /* Last marked as Stale */
    if (cl->lasttime.tv_sec == PRP_SCHEDULE_SIGNAL) isScheduled = TRUE;
    if (cc == not_ready)
    {
      cc = operation_busy;
    }
    else
    { /* Last polled */
      cl->lasttime = tv_stamp;                   
      cl->compStatus = not_signalled;
    }
    cl->starttime.tv_sec = cl->starttime.tv_usec = 0;
    cl->dtimestamp = (SINT32)gServerDataTimeStamp;
    cl->dtimestampMSEC = (int)((gServerDataTimeStamp - (SINT32)gServerDataTimeStamp) * 1000); 
    cl->returnCode = cc;

    for (c=cl->consumer; c != NULL; c=c->next)
    {
      if (c->status.counter == 0)              
      { /* scheduled for removal */
        StaleData++; /* Ensure pass thru doDelivery */
        continue; 
      }
      c->status.statusCode = cc;
      if (isScheduled) c->status.lasttime.tv_sec = c->status.lasttime.tv_usec = PRP_SCHEDULE_SIGNAL;
      if (cc) c->status.Stale = CX_ERROR;      /* Stale because of error code */
      if (c->status.mode == CM_REFRESH || c->status.mode == CM_EVENT)
      {
        if (isScheduled && c->status.mode == CM_EVENT) 
        { /* is an event and was requested as event */
          c->status.Stale = CX_EVENT;
        }
        else if (Stale)                             
        { /* Stale because of data change or system heartbeat */
          c->status.Stale = dirty ? CX_STALE : CX_HEARTBEAT;
          if (dirty && RequireAcknowledgments) c->status.counter = ACK_REQUEST; 
        }
      }
      else /* CM_POLL */
      { /* Honor polling rate   */
        if (MSECS(*tv,c->status.lasttime) > c->status.PollingRate - (SINT32)gSystemTick)
          c->status.Stale = dirty ? CX_STALE : CX_TIMER;/* Stale because of polling rate */
      }
      if (c->status.Stale)                     
      { /* Consumer for this contract is to be refreshed */
        cl->deliveryPending++;
        c->status.lasttime = tv_stamp;
        StaleData++; /* Ensure pass thru doDelivery */
      }
      if (c->status.counter == ACK_REQUEST)    
      { /* Get serious until an acknowledgment comes */
        c->status.mode = ACK_PENDING; /* Reset by renewal */
        c->status.PollingRate = cl->PollingRate;
      }
    }
  }
  gettimeofday(tv,(struct timezone *)NULL);
  for (i=0; StaleData && i<gBurstLimit; i++)
  {
    doDelivery(&thisCycleTime);
  }
  if (StaleData)
  {
    if (gCycleDelay) millisleep(gCycleDelay);
    gBurstLimitReachedCount++;
    if (tineDebug) dbglog("server burst limit %d reached",gBurstLimit);
  }
  ReleaseSystemMutex(hSystemKernelMutex);
  return 0;
}

int prepProducerHeader(int protocol,int xferReason,CONSUMER *cn,CONTRACT *ct,int returnCode,int dstamp,
                       UINT32 tv_sec,UINT32 tv_msec,UINT16 msgsize,BYTE *pPHdrstream)
{
  ProducerHeaderStruct PHdr,*pPHdr = &PHdr;
  ProducerHeaderStruct5 PHdr5,*pPHdr5 = &PHdr5;

  switch (protocol)
  {
    case 5:
      pPHdr5->msgsize = msgsize;
      pPHdr5->CompletionCode = cn->status.statusCode;
      pPHdr5->timestampMSEC = tv_msec;
      pPHdr5->EqmFormat = ct->EqmFormatOut;
      pPHdr5->timestamp = tv_sec;
      pPHdr5->ClnStarttime = cn->status.starttime;
      pPHdr5->SubscriptionID = cn->status.id;
      pPHdr5->counter = (BYTE)cn->status.counter;
      pPHdr5->mtu = cn->status.mtu;
      pPHdr5->blknum = cn->status.blknum;
      pPHdr5->numblks = cn->status.numblks;
#     ifdef INTEL
      memcpy(pPHdrstream,pPHdr5,sizeof(ProducerHeaderStruct5));
#     else
      _StructSwap(ProducerHeaderStruct5Struct,(unsigned char *)pPHdrstream,(unsigned char *)pPHdr5,LTOR);
#     endif
      break;
    case 6:
      pPHdr->msgsize = msgsize;
      pPHdr->CompletionCode = cn->status.statusCode; /* call completion code */
      pPHdr->timestampUSec = tv_msec * 1000;
      pPHdr->EqmFormat = ct->EqmFormatOut;
      pPHdr->timestamp = tv_sec;
      pPHdr->ClnStarttime = cn->status.starttime;
      pPHdr->SubscriptionID = cn->status.id;
      pPHdr->counter = (BYTE)cn->status.counter;
      pPHdr->mtu = cn->status.mtu;
      pPHdr->blknum = cn->status.blknum;
      pPHdr->numblks = cn->status.numblks;
      pPHdr->SrvProtocol = 6;
      pPHdr->xferReason = xferReason;
      pPHdr->stssize = cn->status.statusCode == 0 ? 0 : STATUS_SIZE;
      pPHdr->stscode = returnCode; /* return code of the contract execution */
      pPHdr->userstamp = dstamp;   /* call applied data stamp */
      pPHdr->sysstamp  = gSystemDataStamp;  /* system data stamp */
#     ifdef INTEL
      memcpy(pPHdrstream,pPHdr,sizeof(ProducerHeaderStruct));
#     else
      _StructSwap(ProducerHeaderStructStruct,(unsigned char *)pPHdrstream,(unsigned char *)pPHdr,LTOR);
#     endif
      break;
    default:
      return illegal_protocol;
  }
  return 0;
}

#define maxDeadTime(m,p) ((m) == CM_REFRESH || (m) == CM_EVENT ? HEARTBEAT + (p) : 2 * (p))
int doDelivery(struct timeval *tv)            /* Delivery system */
{
  static int busy = 0;
  CONSUMER *c;
  ContractListStruct *cl;       /* improve readability by assignment */
  short i,k,n,dsize,ssize,fsize,passed,snddata,basemode,actionTaken=FALSE;
  SINT32 extsize, extptr, stssize = 0;
  UINT16 msgsize,totalsize=0,packetmtu;
  UINT32 dptr,delta_t;
  short stepsize,fmt;
  BYTE *buf = srvWorkArea;
  int swallowIt = FALSE, dbgIt;

  if (busy++) return not_ready;
  for (k=0; k<nconsumers; k++)
  {
    if (Consumers[k] == NULL) continue;
    if (Consumers[k]->count == 0)
    {
      RemoveConsumer(k--);
      continue;
    }
    passed = 1; totalsize = sizeof(totalsize);
    for (i=0,n=0; i<ncontracts && n<Consumers[k]->count; i++)
    {
      cl = ContractList[i];
      if (cl->expired == CONTRACT_AVAILABLE) continue;
      for (c=cl->consumer; c != NULL; c=c->next) if (c->client == Consumers[k]) break;
      if (c == NULL) continue;
      if (Consumers[k]->transport == TCP && Consumers[k]->sck == 0)
      { /*  TCP connection error -> remove it */
        c->status.counter = 0;
      }
      n++; snddata = TRUE; extptr = extsize = 0;
      basemode = c->status.mode;
      if (isServiceRequest(&cl->contract) && c->status.statusCode) swallowIt = TRUE;
      if (c->status.Stale)
      {
        actionTaken = TRUE;
        stepsize = PHDR_STEPSIZE(c->client->protocol);
        packetmtu = sizeof(UINT16) + stepsize + c->status.mtu;
        snddata = (c->status.statusCode == 0) ? TRUE : 
                  (c->status.statusCode & CE_SENDDATA) ? TRUE : FALSE;

        stssize = STSSIZE(c->client->protocol);
        if (snddata)
        {
          dsize = (short)c->status.nextDataSize;
          dptr  = c->status.nextDataPtr;
        }
        else
        {
          extsize = stssize;
          dsize = (short)extsize;
          dptr  = 0;
        }
        if ((msgsize=dsize+stepsize) > packetmtu) { c->status.counter = 0; continue; }
        /* if (msgsize % 2 != 0) msgsize++; not necessary ! */ /* make sure of even boundaries */
        prepProducerHeader(c->client->protocol,c->status.Stale,c,&cl->contract,cl->returnCode,cl->datastamp,cl->dtimestamp,cl->dtimestampMSEC,msgsize,&buf[totalsize]);
        if (snddata && (fsize=fmtsizeof(cl->contract.EqmFormatOut)) != 0)
        {
          if ((c->status.statusCode & 0xfff))
          {
            extsize = (SINT32)(c->status.nextDataPtr + c->status.nextDataSize -
                             cl->contract.EqmSizeOut*fmtsizeof(cl->contract.EqmFormatOut));
            if (extsize > stssize) extsize = stssize; /* shouldn't be possible */
            if (extsize > dsize) 
            { /* err string remainder  */
              extsize = dsize;
              extptr  = stssize - extsize;
            }
            if (extsize > 0) dsize -= (short)extsize; else extsize = 0;
          }
          if (dsize > 0)
          {
            memcpy(&buf[totalsize+stepsize],&cl->DataOut[(int)dptr],(size_t)dsize);
            ssize = dsize / fsize; fmt = cl->contract.EqmFormatOut;
            if (c->status.blknum > 1 && /* long CF_SPECTRUM contract */
                LFMT(cl->contract.EqmFormatOut) == CF_SPECTRUM)
            fmt = BFMT(CF_FLOAT); /* remaing data is CF_FLOAT */
            PrepSubData(&buf[totalsize+stepsize],ssize,fmt,LTOR);
          }
        }
        if ((totalsize + msgsize) > packetmtu) { passed = 0; goto bok; }
        totalsize += msgsize;

        if (tineDebug)
        {
          dbgIt = (tineDebug > 2 || c->status.mode <= CM_SINGLE) ? TRUE : FALSE;
          if (tineDebug > 1 && c->status.blknum == c->status.numblks) dbgIt = TRUE;
          if (dbgIt)
            dbglog("[%d] (%s) %s %s <%d>(%d b %d of %d) to %s (+%d)",
              i,cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,
              c->status.statusCode,msgsize,c->status.blknum,c->status.numblks,
              c->client->userName,c->status.counter);
        }
        if (extsize > 0)                      
        { /* insert err string */
          if (c->status.statusCode == server_redirection ||
              c->status.statusCode == call_redirection) 
          {
            if (c->client->protocol < 6)
            {
              memcpy(cl->exp->EqmLastErrorString,&cl->compString[PRP_REDIR_SRV_OFFSET],16);
              memcpy(&cl->exp->EqmLastErrorString[16],&cl->compString[PRP_REDIR_PRP_OFFSET],16);
            }
            else
            {
              memcpy(cl->exp->EqmLastErrorString,cl->compString,STATUS_SIZE);
            }
            c->status.counter = 1;
          }
          else if ((c->status.statusCode & 0xfff) < numErr) 
          {
            SetEqmCompletion(cl->exp->EqmName,erlst[c->status.statusCode & 0xff]);
            if (c->status.statusCode == data_not_local) c->status.counter = 1;
          }
          memcpy(&buf[totalsize-(int)extsize],&cl->exp->EqmLastErrorString[(int)extptr],(size_t)extsize);
        }
        if (snddata == 0 || c->status.numblks == c->status.blknum)
        { /* datablock finished and/or error code */
          c->status.Stale = 0;
          if (c->status.counter > 0 ) c->status.counter--;
          if (c->status.mode == CM_CANCEL) c->status.counter = 0;
          c->status.blknum = 1;
          c->status.nextDataPtr = 0;
          c->status.bytesRemaining = c->status.bytesTotal;
          ReturnToServerCycle = 0;
          cl->deliveryPending--;
          if ((delta_t=MSECS(*tv,c->status.lasttime)) > (UINT32)maxDeadTime(c->status.mode,c->status.PollingRate))
          {
            if (tineDebug > 1)
            {
              dbglog("[%d] (%s) %s %s -> returning late data (%d > %d ms)",
                i,cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,
                (int)delta_t,c->status.PollingRate);
            }
            c->status.misses++;
            gClientMisses++;
          }
        }
        else
        {
          ReturnToServerCycle = c->status.numblks - c->status.blknum;
          c->status.blknum++;
          c->status.nextDataPtr += c->status.nextDataSize;
          c->status.bytesRemaining = c->status.bytesTotal - c->status.nextDataPtr;
          StaleData++;
        }
        c->status.nextDataSize = MIN(c->status.mtu,c->status.bytesRemaining);
      }
      if (c->status.counter == 0) 
      {
        actionTaken = TRUE;
        RemoveConsumerFromList(c,i); 
        n--;
      }
      if (!cl->nconsumers) 
      {
        actionTaken = TRUE;
        if (gRetardSingleContractRemoval && basemode == CM_SINGLE) 
        {
          cl->expired = CONTRACT_EXPIRED;
          cl->contract.EqmAccess &= ~(CA_FIRST|CA_LAST);
          cl->PollingRate = DEFAULT_POLLRATE;
          nExpiredContracts++;
        }
        else 
        {
          RemoveContract(i);
        }
      }
    }

bok:if (totalsize > sizeof(totalsize) && Consumers[k])
    {
      actionTaken = TRUE;
      PrepWord(*(short *)srvWorkArea,totalsize);
      if (!swallowIt) sendToPeer(Consumers[k],srvWorkArea,SRV_RSP_);
    }
    if (!passed) k--;
  }
  gettimeofday(tv,(struct timezone *)NULL);
  busy = 0;
  if (!actionTaken) StaleData = 0;
  return StaleData = --StaleData > 0 ? StaleData : 0;
}
int sendToPeer(CLN *c,BYTE *payload,int msgtype)
{
  switch (c->transport)
  {
    case UDP:
    case TCP:
    case PIPE:
      return sendIPData(c,payload);
#   if defined(IPXNW) || defined(IPXBSD)
    case IPX:
    case SPX:
      return sendIPXData(c,payload);
#   endif
#   if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
    case MMF:
      return sendShmData(c,payload,(WORD)(msgtype == SRV_RSP_ ? WM_RCV_LCL : WM_REQ_LCL));
#   endif
  }
  return illegal_protocol;
}

ContractProgress cp[MAXLONGGRAMS];
void CleanupStrandedDatagrams(void)
{
  int i, n;

  for (i=0; i<MAXLONGGRAMS; i++)
    if (cp[i].busy && thisCycleTime.tv_sec > (SINT32)cp[i].blktime + 5)
  {
    SystemFree((char *)cp[i].data);
    memset(&cp[i],0,sizeof(ContractProgress));
    if (tineDebug > 1) dbglog("stale long datagram %d",i);
  }
  for (i=0, n=0; i<ncontracts; i++)
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;
    if (ContractList[i]->compStatus == not_signalled && ContractList[i]->exp->eqmbusy) n++;
  }
}
int GetDefaultFormatAndSize(CONTRACT *con)
{
  int i, num, idx, cc = illegal_property;
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  char *eqm, *eqpprp;
  BYTE *fmt;
  UINT32 *siz;
  PrpQueryStruct *prpqs;
  XPropertyQueryStruct *xpqs;

  if (con == NULL) ccerr(code_failure);

  eqm = con->EqmName; eqpprp = con->EqmProperty;
  fmt = &con->EqmFormatOut; siz = &con->EqmSizeOut;

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);

  if (el->XEqmPropertyQueryFcn != NULL)        
  { /* registered ext query fcn */
    num = el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs);
    if (num < 0) ccerr(-num);
    for (i=0; i<num; i++) if (!strnicmp(xpqs[i].prpName,eqpprp,PROPERTY_NAME_SIZE))
    {
      *fmt = BFMT(xpqs[i].prpFormat);
      *siz = xpqs[i].prpSize;
      ccerr(0);
    }
  }
  else if (el->EqmPropertyQueryFcn != NULL)         
  { /* registered query fcn */
    num = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
    if (num < 0) ccerr(-num);
    for (i=0; i<num; i++) if (!strnicmp(prpqs[i].prpName,eqpprp,PROPERTY_NAME_SIZE))
    {
      *fmt = BFMT(prpqs[i].prpFormat);
      *siz = prpqs[i].prpSize;
      ccerr(0);
    }
  }
  else                                                    
  { /* handle it here */
    idx = (int)(ElfHash((unsigned char *)eqpprp) % PROPERTY_HASH_SIZE);
    for (prp=el->EqmProperty[idx]; prp; prp=prp->next) if (!strnicmp(prp->prpName,eqpprp,PROPERTY_NAME_SIZE))
    {
      *fmt = BFMT(prp->prpFormat);
      *siz = prp->prpSize;
      ccerr(0);
    }
  }
  if ((i=GetStockPropertyId(eqpprp)) >= 0)
  {
    *fmt = BFMT(StockProperty[i].prpFormat);
    *siz = StockProperty[i].prpSize;
    ccerr(0);
  }
err:
  return cc;
}
int sendMessageToCaller(BYTE *pkthdr,CLN *cln,SUBSCRIPTION *s,int cc,BYTE *data,int len)
{
  BYTE dbuf[HEADERSTRUCT_SIZE+PRODUCER_SIZE+STATUS_SIZE], *buf; /* worst-case buffer size */
  CONSUMER tmpC,*c = &tmpC;
  HeaderStruct tmpH,*h = &tmpH;
  UINT16 msgsize,stssize,psize,subid;
  double dts;

  PrepIncomingHeader(h,pkthdr); /* prepare the packet header */
  switch (cc)
  {
    case get_subscription_id:
     if (h->Protocol < 5) ccerr(illegal_protocol);
     break;
    case call_redirection:
    case server_redirection:
      len = STSSIZE(h->Protocol);
      if (h->Protocol < 6)
      { /* try this and hope for the best */
        memcpy(data,&data[PRP_REDIR_SRV_OFFSET],16);
        memcpy(&data[16],&data[PRP_REDIR_PRP_OFFSET],16);
      }
      break;
    default:
      if (cc > 0 && cc < numErr) 
      {
        if (data == NULL) data = (BYTE *)erlst[cc];
      }
      len = STSSIZE(h->Protocol);
      break;
  }
  if (tineDebug) 
  {
    switch (cc)
    {
      case illegal_property:
      case illegal_equipment_name:
      case non_existent_elem:
      case call_redirection:
      case server_redirection:
      case access_denied:
        dbglog("MSG: (%.8s) %.64s -> %s",s->contract.EqmName,s->contract.EqmProperty,erlst[cc]);
        break;
      case get_subscription_id:
        memcpy(&subid,data,2);
        dbglog("MSG: (id = %d) -> %s",ISWAP(subid),erlst[cc]);
        break;
      default: 
        dbglog("MSG: -> %s",erlst[cc]);
        break;
    }
  }
  (c = &tmpC)->status.starttime = s->starttime;
  c->status.statusCode = cc;
  c->status.id = s->id;
  c->status.numblks = 1;
  c->client = cln;
  c->client->protocol = h->Protocol;
  dts = PutDataTimeStamp(gDataTimeStampOffset,thisCycleTime.tv_sec,(int)(thisCycleTime.tv_usec/1000));
  stssize = STSSIZE(h->Protocol);
  psize = PHDR_STEPSIZE(h->Protocol);
  msgsize = psize + stssize;
  buf = dbuf;
  PrepWord(*(short *)buf,msgsize+sizeof(msgsize));
  prepProducerHeader(h->Protocol,CX_MESSAGE,c,&s->contract,not_posted,0,(UINT32)dts,(int)((dts - (SINT32)dts) * 1000),msgsize,&buf[sizeof(msgsize)]);
  
  if (data != NULL && len > 0)
  {
    memcpy(&buf[sizeof(msgsize)+psize],data,len);
  }
  sendToPeer(cln,buf,SRV_RSP_);
err:
  return cc;
}
int getNextFreeContractID(void)
{
  static UINT16 cid = 0;
  int i;
  for (i=0,cid++; i<ncontracts; i++) 
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;
    if (ContractList[i]->contractID == cid) cid++;
  }
  return cid;
}
#define contract5to6(c6,c5) {\
  strncpy(c6.EqmProperty,c5.EqmProperty,PROPERTY_NAME_SHORTSIZE);\
  strncpy(c6.EqmDeviceName,c5.EqmDeviceName,DEVICE_NAME_SHORTSIZE);\
  strncpy(c6.EqmName,c5.EqmName,EQM_NAME_SIZE);\
  c6.EqmSizeIn = c5.EqmSizeIn;\
  c6.EqmSizeOut = c5.EqmSizeOut;\
  c6.hEqmName = c5.hEqmName;\
  c6.EqmAccess = c5.EqmAccess;\
  c6.EqmFormatIn = c5.EqmFormatIn;\
  c6.EqmFormatOut = c5.EqmFormatOut;\
  strncpy(c6.strTagIn,c5.strTagIn,TAG_NAME_SHORTSIZE);\
  strncpy(c6.strTagOut,c5.strTagOut,TAG_NAME_SHORTSIZE);\
}
int prepSubHeader(int protocol,SUBSCRIPTION *pSub,BYTE *pSubstream)
{
  SUBSCRIPTION5 sub5, *pSub5 = &sub5;
  switch (protocol)
  {
    case 5:
#     ifdef INTEL
      if (ForceStructAlignment)
      { /* NIOS for instance */
        memcpy(pSub5,pSubstream,sizeof(SUBSCRIPTION));
      }
      else
      { /* linux, windows, dos, etc. */
        pSub5 = (SUBSCRIPTION5 *)pSubstream;
      }
#     else
      memset(pSub5,0,sizeof(SUBSCRIPTION5));
      _StructSwap(SUBSCRIPTION5Struct,pSubstream,(BYTE *)pSub5,RTOL);
#     endif
      memset(pSub,0,sizeof(SUBSCRIPTION));
      pSub->msgsize = pSub5->msgsize;
      pSub->mtu = pSub5->mtu;
      pSub->numblks = pSub5->numblks;
      pSub->blknum = pSub5->blknum;
      pSub->id = pSub5->id;
      pSub->mode = pSub5->mode;
      pSub->blkid = pSub5->blkid;
      pSub->PollingRate = pSub5->PollingRate;
      pSub->starttime = pSub5->starttime;
      contract5to6(pSub->contract,pSub5->contract);
      if (pSub->contract.EqmDeviceName[DEVICE_NAME_SHORTSIZE-1] == '&')
      { /* legacy kluge to support channel access names as device names */
        pSub->contract.EqmDeviceName[DEVICE_NAME_SHORTSIZE-1] = 0;
        strncat(pSub->contract.EqmDeviceName,&pSub->contract.EqmProperty[DEVICE_NAME_SHORTSIZE+1],DEVICE_NAME_SHORTSIZE-1);
        memset(&pSub->contract.EqmProperty[DEVICE_NAME_SHORTSIZE],0,DEVICE_NAME_SHORTSIZE);
      }
      break;
    case 6:
#     ifdef INTEL
      memcpy(pSub,pSubstream,sizeof(SUBSCRIPTION));
#     else
      memset(pSub,0,sizeof(SUBSCRIPTION));
      _StructSwap(SUBSCRIPTIONStruct,pSubstream,(BYTE *)pSub,RTOL);
#     endif
      break;
    default:
       return illegal_protocol;
  }
  return 0;
}
int isQueryRequest(CONTRACT *con)
{
  int i;
  if (con == NULL) return FALSE;
  if (GetStockPropertyId(con->EqmProperty) >= 0) return TRUE;
  for (i=0; i<NrStockModifiers; i++)
  { /* run through the list of stock modifiers */
    if (strstr(con->EqmProperty,StockModifier[i].modName)) return TRUE;
  }
  return FALSE;
}
int contractsEqual(CONTRACT *c1, CONTRACT *c2)
{
  if (c1 == NULL || c2 == NULL) return FALSE;
  if (memcmp(c1,c2,PROPERTY_NAME_SIZE+DEVICE_NAME_SIZE+EQM_NAME_SIZE+2*sizeof(UINT32))) return FALSE;
  if ((c1->EqmAccess & 0x03) != (c2->EqmAccess & 0x03)) return FALSE;
  if (c1->EqmFormatIn != c2->EqmFormatIn) return FALSE;
  if (c1->EqmFormatOut != c2->EqmFormatOut) return FALSE;
  if (strncmp(c1->strTagIn,c2->strTagIn,TAG_NAME_SIZE)) return FALSE;
  if (strncmp(c1->strTagOut,c2->strTagOut,TAG_NAME_SIZE)) return FALSE;
  return TRUE;
}
static char *getContractState(int state)
{
  switch (state)
  {
    case CONTRACT_ACTIVE: return "active";
    case CONTRACT_EXPIRED: return "expired";
    case CONTRACT_AVAILABLE: return "available";
    default: return "unknown";
  }
}
#define PAYLOAD_OFFSET(p) (bytesread + ((p) == 5 ? SUBSCRIPTION5_SIZE : SUBSCRIPTION_SIZE + sub.extsize))
#define XSTRINGSET_OFFSET SUBSCRIPTION_SIZE
UINT32 gSingleLinkCount = 0;
#define replaceContractAliasNames(cn) { if (nAliasTableEntries > 0) {\
  if ((i=getAliasTableIndex((cn)->EqmProperty)) >= 0) strncpy((cn)->EqmProperty,gAliasTable[i].name,PROPERTY_NAME_SIZE);\
  if ((i=getAliasTableIndex((cn)->EqmDeviceName)) >= 0) strncpy((cn)->EqmDeviceName,gAliasTable[i].name,DEVICE_NAME_SIZE);\
  }\
}
#define ccout(o) { cc = o; goto out; }
int getConsumerRequest(CLN *caller,BYTE *buf,int packetsize)
{
  BYTE *dbuf;
  HeaderStruct tmpH,*h = &tmpH;
  CONSUMER tmpC,*c = &tmpC;
  CLN *nwcln, *cln=caller, newcln;
  int i,j = -1, slot = -1;
  short subid,fmtinsize,fmtoutsize,cc=0,isTrueMCast=0,renewalMultiplier,swallowOnErr=FALSE;
  UINT16 totalsize,msgsize,datasize,bytesread=HEADERSTRUCT_SIZE;
  UINT32 dinsize,doutsize,dinallocsize=0,defsize,dataPtr;
  SUBSCRIPTION sub,*pSub = &sub;
  UINT16 datamtu;
  int  extsiz = 0;
  char extstr[DEVICE_NAME_MAXSIZE],stsstr[STATUS_SIZE];

  stsstr[0] = 0;
  msgsize = 0; /* avoid compiler warning */
  PrepIncomingHeader(h,buf); totalsize = h->TotalSize;

  if (totalsize <= 0) ccout(out_of_range);

  if (tineDebug) sprintf(&dbgbuf[strlen(dbgbuf)],"%s -> ",h->UserName);
  strncpy(cln->userName,h->UserName,15);                  /* always reset   */

  /* map incoming protocol to the most modern : at the moment 6 */
  if ((cc=prepSubHeader(h->Protocol,&sub,&buf[bytesread])) != 0) ccout(cc);
  if (sub.extsize > 0 && sub.blknum == 1)
  {  /* extended (device name) string set (comes with first block) */
     extsiz = MIN(sub.extsize,DEVICE_NAME_MAXSIZE);
     memcpy(extstr,&buf[bytesread+XSTRINGSET_OFFSET],extsiz);
     if (extsiz % 2) extsiz++;
  }

  replaceContractAliasNames(&pSub->contract);
  if (isServiceRequest(&pSub->contract)) swallowOnErr = TRUE;
  while (bytesread < totalsize && bytesread < packetsize) /* unpack request */
  {
    stsstr[0] = 0;
    isTrueMCast = FALSE;
    cln = caller; /* reset this in case of packed NETWORK request ! */
    if (tineDebug && pSub->numblks < 4) /* output to screen can influence io buffering */
    {
      sprintf(&dbgbuf[strlen(dbgbuf)],"%s %s %s %d msec (%d of %d)[%s]",
              pSub->contract.EqmName,pSub->contract.EqmDeviceName,
              pSub->contract.EqmProperty,(int)pSub->PollingRate,
              pSub->blknum,pSub->numblks,SubType[(pSub->mode)&0x03]);
      dbglog(dbgbuf);
      sprintf(dbgbuf,"\t\t ->");
    }
    if (NrOfDeniedUsers > 0 && isDeniedUser(h->UserName)) ccerr(access_denied);
    if ((fmtoutsize=fmtsizeof(pSub->contract.EqmFormatOut)) > 0)
    {
      if ((doutsize=pSub->contract.EqmSizeOut*fmtoutsize) <= 0) ccerr(illegal_data_size);
      if (LFMT(pSub->contract.EqmFormatOut) == CF_SPECTRUM) doutsize += 96; /* CF_SPECTRUM HEADER */
      if (doutsize > (UINT32)MaxSystemTransportSize) 
      {
        if (!isQueryRequest(&pSub->contract)) ccerr(invalid_transport_size);
        pSub->contract.EqmSizeOut = MaxSystemTransportSize / fmtoutsize;
        doutsize = pSub->contract.EqmSizeOut*fmtoutsize;
      }
    }
    else doutsize = 0;

    if ((fmtinsize=fmtsizeof(pSub->contract.EqmFormatIn)) != 0) 
    { /* data input */
      if (pSub->blknum > pSub->numblks) ccerr(link_error);
      dinsize = pSub->contract.EqmSizeIn*fmtinsize;
      if (LFMT(pSub->contract.EqmFormatIn) == CF_SPECTRUM) dinsize += 96; /* CF_SPECTRUM HEADER */
      dbuf = &buf[PAYLOAD_OFFSET(h->Protocol)];
      dataPtr = ((UINT32)(pSub->blknum-1))*pSub->mtu;
      datasize = MIN((UINT16)(dinsize-dataPtr),pSub->mtu);
      if (dinsize > (UINT32)MaxSystemTransportSize) ccerr(invalid_transport_size);
      dinallocsize = ALLOCSIZE(dinsize);
      if (dinsize > pSub->mtu) /* incoming long datagram */
      {
        for (i=0; i<MAXLONGGRAMS; i++)
          if (cp[i].haddr == (UINT32)cln->IPaddress.sin_addr.s_addr &&
              cp[i].SubscriptionID == pSub->id &&
              cp[i].SubsStarttime  == pSub->starttime) break;
        if (i == MAXLONGGRAMS) for (i=0; cp[i].busy && i<MAXLONGGRAMS; i++);
        if (i == MAXLONGGRAMS) ccerr(resources_exhausted);
        if (++(cp[i].blksin) <= pSub->numblks)
        {
          if (cp[i].blksin == 1) cp[i].blkid = pSub->blkid;
          else if (cp[i].blkid < pSub->blkid ||
                   cp[i].blkid > pSub->blkid + 500)
          {
            if (tineDebug > 1)
              dbglog("re-sync long datagram id %d to %d",cp[i].blkid,pSub->blkid);
            cp[i].blksin = 1;
            cp[i].blkid = pSub->blkid;
            cp[i].blktime = (UINT32)time(NULL);
            cp[i].SubsStarttime = pSub->starttime;
          }
          if (!cp[i].busy)
          {
            if (tineDebug > 1) dbglog("new long datagram (%d) sub id %d %d bytes",i,pSub->id,(int)dinsize);
            cp[i].busy = 1;
            cp[i].haddr = cln->IPaddress.sin_addr.s_addr;
            cp[i].SubscriptionID = pSub->id;
            cp[i].blktime = (UINT32)time(NULL);
            cp[i].SubsStarttime = pSub->starttime;
            if ((cp[i].data=(BYTE *)SystemCalloc(1,(size_t)dinallocsize)) == NULL) ccerr(out_of_server_memory);
          }
          memcpy(&cp[i].data[(int)dataPtr],dbuf,(size_t)datasize);
          ReturnToServerCycle = pSub->numblks - cp[i].blksin;
          if (cp[i].blksin < pSub->numblks) goto bow;
          cp[i].busy = 2;
          dbuf = cp[i].data;
          if (tineDebug > 1) dbglog("in-coming long datagram (%d blocks) complete",pSub->numblks);
        }
        else ReturnToServerCycle = 0;
      }
      PrepSubData(dbuf,pSub->contract.EqmSizeIn,pSub->contract.EqmFormatIn,RTOL);
    }
    else 
    { /* no input data -> make sure the input tag is zeroed out */
      memset(pSub->contract.strTagIn,0,TAG_NAME_SIZE);
      pSub->contract.EqmSizeIn = 0;
      dinsize = 0; dinallocsize = 0; 
      dbuf = NULL;
    }

    if (pSub->PollingRate < MinPollingRate) pSub->PollingRate = MinPollingRate;
    if (pSub->PollingRate > 1000) pSub->PollingRate = (pSub->PollingRate/1000)*1000;

    if (LFMT(pSub->contract.EqmFormatOut) == CF_DEFAULT)
    { /* caller wants default settings */
      if ((cc=GetDefaultFormatAndSize(&pSub->contract)) != 0) ccerr(cc);
      if ((fmtoutsize=fmtsizeof(pSub->contract.EqmFormatOut)) != 0)
      {
        if ((defsize=pSub->contract.EqmSizeOut*fmtoutsize) <= 0) ccerr(illegal_data_size);
        if (LFMT(pSub->contract.EqmFormatOut) == CF_SPECTRUM) defsize += 96; /* CF_SPECTRUM HEADER */
        if (defsize > (UINT32)MaxSystemTransportSize) ccerr(invalid_transport_size);
        if (defsize > doutsize)           
        { /* caller can't buffer default */
          pSub->contract.EqmSizeOut = doutsize/fmtoutsize;
          defsize = pSub->contract.EqmSizeOut*fmtoutsize;
          if (LFMT(pSub->contract.EqmFormatOut) == CF_SPECTRUM) 
          { /* subtract CF_SPECTRUM HEADER */
            if (pSub->contract.EqmSizeOut > 24) pSub->contract.EqmSizeOut -= 24; 
          }
        }
      }
      else defsize = 0;
      doutsize = defsize; /* for allocation purposes */
    }
    slot = -1; /* start search for available slot */
    for (j=0; j<ncontracts; j++)            
    { /* is contract in list ? */
      if (ContractList[j]->expired == CONTRACT_AVAILABLE) 
      {
        if (slot < 0) slot = j;
        continue;
      }
      if (!memcmp(&ContractList[j]->contract,&pSub->contract,sizeof(CONTRACT)))
      {
        if (extsiz > 0)
        { /* extended string coming */
          if (ContractList[j]->extStringSpace == NULL || 
              strncmp(ContractList[j]->extStringSpace,extstr,extsiz)) continue;
        }
        if (!dbuf && !ContractList[j]->DataIn) break; /* no data input */
        if (dbuf && ContractList[j]->DataIn) 
        { /* input buffer has been allocated */
          if (!memcmp(ContractList[j]->DataIn,dbuf,(size_t)dinsize)) break;
        }
      }
    }
    if (j == ncontracts)                    
    { /* no, add to list  */
      if (slot >= 0) 
      {
        j = slot;
        memset(ContractList[j],0,sizeof(ContractListStruct));
        if (nExpiredContracts > 0) nExpiredContracts--;
      }
      else
      {
        if (j >= MaxNumContracts) ccerr(resources_exhausted);
        if ((ContractList[j]=(ContractListStruct *)SystemCalloc(1,sizeof(ContractListStruct))) == NULL)
          ccerr(out_of_server_memory);
        ncontracts++;
      }
      memcpy(&ContractList[j]->contract,&pSub->contract,sizeof(CONTRACT));
      if (extsiz > 0)
      {
        if ((ContractList[j]->extStringSpace = (char *)SystemCalloc(1,DEVICE_NAME_MAXSIZE)) == NULL)
          ccerr(out_of_server_memory);
        ContractList[j]->extStringSize = extsiz;
        strncpy(ContractList[j]->extStringSpace,extstr,extsiz);
      }
      ContractList[j]->compStatus = not_signalled;
      ContractList[j]->returnCode = not_posted;
      ContractList[j]->listId = j;

      if ((cc=IsIllegalProperty(ContractList[j]->contract.EqmName,
                                ContractList[j]->contract.EqmProperty,
                                ContractList[j]->contract.EqmDeviceName,
                                cln,
                                ContractList[j]->contract.EqmAccess,
                                &ContractList[j]->exp,
                                stsstr)) != 0) ccerr(cc);

      if (dinallocsize && (ContractList[j]->DataIn = (BYTE *)SystemCalloc(1,(size_t)dinallocsize)) == NULL)
      {
        ccerr(out_of_server_memory);
      }
      if (doutsize && (ContractList[j]->DataOut = (BYTE *)SystemCalloc(nDataOutBuffers,(size_t)doutsize)) == NULL)
      { /* multi-threaded => nDataOutBuffers = 2, otherwise srvWorkArea used for calls */
        if (dinsize) SystemFree((char *)ContractList[j]->DataIn);
        ccerr(out_of_server_memory);
      }
      ContractList[j]->DataBuf = nDataOutBuffers > 1 ? &ContractList[j]->DataOut[doutsize] : srvWorkArea;
      if (dbuf) memcpy(ContractList[j]->DataIn,dbuf,(size_t)dinsize);
      ContractList[j]->contract.EqmAccess |= CA_FIRST; /* turn on FIRST bit */
      ContractList[j]->contractID = getNextFreeContractID();
      ContractList[j]->expired = CONTRACT_ACTIVE;
      ContractList[j]->PollingRate = pSub->PollingRate; /* reset this too ! */
      ContractList[j]->lasttime.tv_sec = PRP_SCHEDULE_SIGNAL;
      if (tineDebug > 1)
      {
        dbglog("contract [%d] (%s) %s %s created",
          j,pSub->contract.EqmName,pSub->contract.EqmDeviceName,pSub->contract.EqmProperty);
      }
    }
    else
    { /* contract was found in the list */
      if (tineDebug > 1)
      {
        dbglog("contract [%d] (%s) %s %s (%s) attached",
          j,pSub->contract.EqmName,pSub->contract.EqmDeviceName,pSub->contract.EqmProperty,
          getContractState(ContractList[j]->expired));
      }
    }
    if (ContractList[j]->expired == CONTRACT_EXPIRED && !(pSub->mode & CM_RETRY)) 
    { /* is a new link attaching to an expired contract ? */
      ContractList[j]->expired = CONTRACT_ACTIVE;
      ContractList[j]->contract.EqmAccess |= CA_FIRST; /* turn on FIRST bit */
      if (nExpiredContracts > 0) nExpiredContracts--;
      if (tineDebug > 1) dbglog("re-activating contract");
    }
#   if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
    if (cln->transport == MMF && pSub->mode & (CM_BCAST|CM_MCAST)) 
    { /* a local client is asking for a network subscription */
      if (gForceShmDatagrams || getNetworkClientFromList() == NULL)
      { /* there's no network client going at the moment */ 
        pSub->mode &= ~(CM_BCAST|CM_MCAST); /* strip this off for a more efficient MMF transfer */
      }
    }
#   endif
    if (pSub->mode & (CM_BCAST|CM_MCAST))
    { /* is broadcast or multicast request ? */
      isTrueMCast = pSub->mode & CM_MCAST;
      pSub->mode &= 0x0ff; /* strip off */
      if (pSub->mode == CM_CANCEL) ccerr(0); /* don't do anything */
      if (IsMemberBcastNets(cln) || isTrueMCast)
      { /* send subscription ID to caller */
        subid = ISWAP(ContractList[j]->contractID);
        sendMessageToCaller(buf,cln,pSub,get_subscription_id,(BYTE *)&subid,2);
        if ((nwcln=getNetworkClientFromList()) != NULL) 
        { /* already servicing the NETWORK client */
          cln = nwcln;
          if (tineDebug > 1) dbglog("additional NETWORK client attaching");
        }
        else
        { /* Network client not yet in list */
          memcpy(&newcln,cln,sizeof(CLN)); /* don't dork the existing settings */
          cln = &newcln;
          if (cln->transport == PIPE || cln->transport == MMF)
          { 
            cln->transport = UDP; 
            cln->IPaddress.sin_family = AF_INET;
            cln->IPaddress.sin_addr.s_addr = tMCastAddr.s_addr;
            cln->sck = ProducerSocket; /* reset this in case of windows MMF */
          }
          if (cln->IPaddress.sin_addr.s_addr != 0)
          { /* add network to client list and replace cln */
            cln->IPaddress.sin_port = htons((UINT16)gtMCastPort);
            if (isTrueMCast) cln->IPaddress.sin_addr.s_addr = tMCastAddr.s_addr;
            else cln->IPaddress.sin_addr.s_addr |= 0xff000000L;
            cln = LocateConsumerInList(&cln->IPaddress,NULL,NULL);
            if (tineDebug > 1) dbglog("add NETWORK to client list");
          }
          else if (*(UINT32 *)cln->IPXaddress.network != 0)
          { /* IPX Client ! -> add network to client list and replace cln */
            memset(&cln->IPXaddress.network[10],MCAST_SOCKET,2);
            memset(cln->IPXaddress.node,0xff,6);
            cln = LocateConsumerInList(NULL,&cln->IPXaddress,cln->node);
          }
          cln->sck = newcln.sck; cln->transport = newcln.transport;
          cln->protocol = newcln.protocol;
          strncpy(cln->userName,"NETWORK",USERNAME_SIZE);
        }
        pSub->id = ContractList[j]->contractID;
        pSub->starttime = BCAST_ID;
      }
    }

    /* is client in contractList?    */
    for (c=ContractList[j]->consumer; c!=NULL && c->client!=cln; c=c->next);

    if (c == NULL)  
    { /* no, add to list */
      if ((c=(CONSUMER*)SystemCalloc(1,sizeof(CONSUMER))) == NULL) ccerr(out_of_server_memory);
      c->next = ContractList[j]->consumer;
      ContractList[j]->consumer = c;
      c->client = cln;
      ContractList[j]->nconsumers++;
      c->client->count++;
      c->client->protocol = h->Protocol;
      c->status.statusCode = ContractList[j]->returnCode;
      c->status.counter = 1;
      if (tineDebug > 1) dbglog("append client %s to contract %d",cln->userName,j);
    }
    c->status.PollingRate = pSub->PollingRate;
    c->status.id = pSub->id;
    c->status.starttime = pSub->starttime;
    c->status.blknum = 1;
    c->status.nextDataPtr = 0;
    c->status.bytesRemaining = doutsize;
    c->status.bytesTotal = doutsize;
    datamtu = PMTU(c->client->transport,h->Protocol);
    if (fmtoutsize > 0) datamtu -= datamtu%fmtoutsize;
    c->status.mtu = datamtu;
    c->status.nextDataSize = c->status.bytesRemaining;
    if (c->status.bytesRemaining > c->status.mtu) c->status.nextDataSize = c->status.mtu;
    c->status.numblks = doutsize/((UINT32)c->status.mtu) + (doutsize%((UINT32)c->status.mtu) ? 1 : 0);
    if (doutsize == 0) c->status.numblks = 1;
    renewalMultiplier = 1;
    if (pSub->PollingRate < 500) renewalMultiplier = 3;
    if (c->status.mode != ACK_PENDING)
    { /* Force Contract execution */
      if (pSub->starttime == BCAST_ID)
      { /* latching on the a NETWORK contract */
        if ((ContractList[j]->contract.EqmAccess & CA_FIRST) == 0)
        { /* wait for next delivery cycle */
          if (tineDebug > 1) dbglog("additional network client to contract");
          c->status.lasttime = thisCycleTime;
          if (ContractList[j]->PollingRate > 1000) c->status.lasttime.tv_sec -= ContractList[j]->PollingRate;
        }
        else
        {
          if (tineDebug > 1) dbglog("initial network client for contract");
        }
      }
      else if (c->status.counter == 0 || pSub->mode == CM_REFRESH || pSub->mode == CM_EVENT)
      { /* send out the last acquired data */
        if (ContractList[j]->returnCode != not_posted) 
        { /* contract return code can be trusted */
          c->status.Stale = CX_ATTACH; 
          c->status.lasttime = thisCycleTime;
        }
      }
      if (pSub->mode != CM_CANCEL && pSub->mode != CM_REFRESH && pSub->mode != CM_EVENT &&
          c->status.counter > ACK_REQUEST*renewalMultiplier &&
          c->status.counter < RENEWAL*renewalMultiplier)
      {
        if (tineDebug > 1)
        {
          dbglog("reconnecting to %s %s %s from %s (+%d)",
            pSub->contract.EqmName,pSub->contract.EqmDeviceName,
            pSub->contract.EqmProperty,h->UserName,c->status.counter);
        }
        c->status.reconnects++;
        gClientReconnects++;
      }
      StaleData++;
    }
    if (pSub->mode & CM_RETRY)
    {
      if (tineDebug > 1)
      {
        dbglog("retrying %s %s %s by %s",
          pSub->contract.EqmName,pSub->contract.EqmDeviceName,
          pSub->contract.EqmProperty,h->UserName);
      }
      c->status.Stale = CX_RETRY;                     
      StaleData++;   /* if expired, doDelivery() will send the last payload again */
      c->status.isRetry = TRUE;
      gClientRetries++;
      pSub->mode = BASEMODE(pSub->mode);
    }
    if (ContractList[j]->nconsumers == 0)       
    { /* just turned off, turn it back on */
      if (tineDebug) 
      {
        dbglog("re-activate contract %s %s %s",
          pSub->contract.EqmName,pSub->contract.EqmDeviceName,pSub->contract.EqmProperty);
      }
      ContractList[j]->nconsumers++;
      ContractList[j]->contract.EqmAccess &= ~(CA_LAST);
    }
    switch (c->status.mode = pSub->mode)
    {
      case CM_SINGLE :
        gSingleLinkCount++;
        c->status.counter = 1;
        if (!(ContractList[j]->expired == CONTRACT_EXPIRED && c->status.isRetry))
        {
          ContractList[j]->lasttime.tv_sec = PRP_RETRY_SIGNAL;  
        }
        else  
        { /* a RETRY marked as expired should not be called */
          ContractList[j]->lasttime.tv_sec = (UINT32)time(NULL);
          if (ContractList[j]->dataSizeOut != ContractList[j]->contract.EqmSizeOut)
            renegotiateContract(ContractList[j],ContractList[j]->returnCode);
        }
        if (ContractList[j]->PollingRate == 0) ContractList[j]->PollingRate = c->status.PollingRate;
        if (ContractList[j]->nconsumers == 1)
        {
          ContractList[j]->contract.EqmAccess |= CA_LAST; /* turn on LAST bit */
        }
        break;
      case CM_POLL   :
      case CM_REFRESH:
      case CM_EVENT:
        c->status.counter = RENEWAL * renewalMultiplier;
        ContractList[j]->PollingRate = findpoll(pSub->PollingRate,ContractList[j]->PollingRate);
        break;
      default     :
      case CM_CANCEL :
        if (!strcmp(c->client->userName,"NETWORK")) break;
        if (ContractList[j]->nconsumers == 1)
        {
          if (c->status.counter) 
          { /* was active : turn on LAST bit */
            ContractList[j]->contract.EqmAccess |= CA_LAST; 
          }
          ContractList[j]->lasttime.tv_sec = PRP_CANCEL_SIGNAL;
          ContractList[j]->deliveryPending = 0;
        }
        c->status.counter = 0;
        break;
    }
    if (tineDebug > 2)
      dbglog("link counter set to %d for %s (mode %d)",c->status.counter,c->client->userName,c->status.mode);
err:
    if (cc)
    {
      if (cc != server_redirection && strlen(stsstr) == 0) strncpy(stsstr,erlst[cc],32);
      if (!swallowOnErr) sendMessageToCaller(buf,cln,pSub,cc,(BYTE *)stsstr,STATUS_SIZE);
      if (j >= 0 && j < MaxNumContracts) RemoveContract(j);
      cc = 0;
    }
    else
    {
      SysPoll = findpoll(SysPoll,ContractList[j]->PollingRate);
    }
    for (i=0; i<MAXLONGGRAMS; i++) if (cp[i].busy == 2)
    {
      SystemFree((char *)cp[i].data);
      memset(&cp[i],0,sizeof(ContractProgress));
      if (tineDebug > 1) dbglog("discarding long datagram %d",i);
    }
bow:
    if (pSub->msgsize <= 0) 
    { /* appears to be super corrupt ! */
      if (tineDebug) dbglog("corrupt message size : %d",pSub->msgsize);
      break; 
    }
    if ((bytesread += pSub->msgsize) < totalsize)
    {
      if (pSub->msgsize % 2 != 0) break;
      if ((cc=prepSubHeader(h->Protocol,&sub,&buf[bytesread])) != 0) ccout(cc);
      extsiz = 0;
      if (sub.extsize > 0)
      {  /* extended (device name) string set */
         extsiz = MIN(sub.extsize,DEVICE_NAME_MAXSIZE);
         memcpy(extstr,&buf[XSTRINGSET_OFFSET],extsiz);
         if (extsiz % 2) extsiz++;
      }
      replaceContractAliasNames(&pSub->contract);
    }
  } /* continue unpacking  */
out:
  if (cc && !swallowOnErr) 
  {
    if (strlen(stsstr) == 0) strncpy(stsstr,erlst[cc],32);
    sendMessageToCaller(buf,cln,pSub,cc,(BYTE *)stsstr,STATUS_SIZE);
  }
  return cc;
}
void doGCast(struct timeval *thisTime)
{
  BYTE *buf = srvWorkArea;
  int cc = 0,i,n,setsize,GlobalPacketSize,datasize;
  char *strptr;
  GlobalDataHdr tmpBD,*gd = &tmpBD;
  HeaderStruct tmpH,*GlobalHdr = &tmpH;
  int delta_t;
  time_t interim = thisTime->tv_sec;
  static int done = FALSE;

  if (!done)
  {
    if (initSendNetGlobalsList()) return;
    done = TRUE;
  }
  if (WaitForSystemMutex(hSystemKernelMutex,gSystemTick) != 0) return;
  GlobalPacketSize = HEADERSTRUCT_SIZE;
  for (i=0, n=0; i<ngcasts; i++)
  {
    if (gcastTbl[i]->tv.tv_sec > 0 &&                         /* not a scheduled broadcast */
        (gcastTbl[i]->minPeriod <= 0 || gcastTbl[i]->maxPeriod <= 0)) continue;
    delta_t = (int)MSECS(*thisTime,gcastTbl[i]->tv);
    if (delta_t < gcastTbl[i]->minPeriod) continue; /* too soon */
    if (delta_t < gcastTbl[i]->maxPeriod) /* within heartbeat */
    {
      if (!datacmp(gcastTbl[i]->refdata,
            gcastTbl[i]->data,
            gcastTbl[i]->size,
            gcastTbl[i]->format,
            gcastTbl[i]->tolerance)) continue;
    }
    datasize = gcastTbl[i]->size*fmtsizeof(gcastTbl[i]->format);
    setsize = GLOBALDATA_HEADER_SIZE + datasize;
    if (GlobalPacketSize + setsize > gBrdcstBufferSize)
    {                                          /* send what's there  */
      AssignOutgoingHeader(GlobalHdr,buf);  /* fill in the header */
      GlobalHdr->EqmSize = n;               /* number of keywords */
      strncpy(GlobalHdr->UserName,gUserName,USERNAME_SIZE);
      GlobalHdr->Protocol = 3;              /* broadcast protocol */
      GlobalHdr->TotalSize = GlobalPacketSize;
      PrepOutgoingHeader(GlobalHdr,buf);
      cc = sendGCast(buf,GlobalPacketSize);
      GlobalPacketSize = HEADERSTRUCT_SIZE;
      n = 0;
    }
    AssignOutgoingBrdcstData(gd,&buf[GlobalPacketSize]);
    strncpy(gd->name,gcastTbl[i]->keyword,16);
    gd->size = gcastTbl[i]->size;
    gd->format = gcastTbl[i]->format;
    gd->time = (SINT32)(gcastTbl[i]->time = interim);
    gcastTbl[i]->tv = *thisTime;
    PrepOutgoingGCastData(gd,&buf[GlobalPacketSize]);
    GlobalPacketSize += GLOBALDATA_HEADER_SIZE;
    if (gcastTbl[i]->data)
    {
      memcpy(gcastTbl[i]->refdata,gcastTbl[i]->data,datasize);
      PrepOutgoingData(&buf[GlobalPacketSize],gcastTbl[i]->data,datasize,gcastTbl[i]->format);
    }
    if (tineDebug > 1)
    {
      sprintf(dbgbuf,"GCAST: %s ",gcastTbl[i]->keyword);
      strptr = &dbgbuf[strlen(dbgbuf)];
      if (cc) sprintf(strptr,"(ERR: %s)",erlst[cc]);
      dbglog(dbgbuf);
      if (cc == 0 && tineDebug > 2)
      {
        DataDump("GCAST: ",(char *)gcastTbl[i]->data,
                  gcastTbl[i]->size,gcastTbl[i]->format,LTOR);
      }
    }
    if (gcastTbl[i]->fcn) (gcastTbl[i]->fcn)(i,cc);
    GlobalPacketSize += datasize; n++;
  }
  if (n > 0)                                 /* something left     */
  {
    AssignOutgoingHeader(GlobalHdr,buf);  /* fill in the header */
    GlobalHdr->EqmSize = n;               /* number of keywords */
    strncpy(GlobalHdr->UserName,gUserName,USERNAME_SIZE);
    GlobalHdr->Protocol = 3;              /* broadcast protocol */
    GlobalHdr->TotalSize = GlobalPacketSize;
    PrepOutgoingHeader(GlobalHdr,buf);
    cc = sendGCast(buf,GlobalPacketSize);
  }
  ReleaseSystemMutex(hSystemKernelMutex);
  return;
}

char tExecLinkReturnedCompletion[STATUS_SIZE+4];
int tExecLinkReturnedDataSize = 0;
int gTineServicesProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
int tExecLink(char *fec,char *eqm,char *dev,char *prp,DTYPE *dout,DTYPE *din,short access,short timout)
{
  int fecindex,fmtsize,fmtsizeIn,retry = 2;
  short siz=0,fmt=CF_NULL,sizIn=0,fmtIn=CF_NULL;
  char *datOut=NULL,*datIn=NULL,*stagOut=NULL,*stagIn=NULL;
  short cc=0,dsize,dsizeIn,dretsiz=0,sstepsize,pstepsize;
  static BYTE lclWorkArea[IPX_BUFFER_SIZE];
  static UINT16 tid = 0;            /* trivial subscription id */
  HeaderStruct hTmp,*h = &hTmp;
  ProducerHeaderStruct PHdr, *pPHdr = &PHdr;
  ConnTableEntry c;

  if (tineDebug) dbglog("trivial RPC: %s@%s %s %s",eqm,fec,prp,dev);
  if ((fecindex=getFecIndex(fec)) < 0) return -fecindex;
  if (dout != NULL)
  {
    fmt = dout->dFormat;
    siz = dout->dArrayLength;
    datOut = dout->data.cptr;
    stagOut = dout->dTag;
  }
  if (din != NULL)
  {
    fmtIn = din->dFormat;
    sizIn = din->dArrayLength;
    datIn = din->data.cptr;
    stagIn = din->dTag;
  }
  fmtsize = fmtsizeof(fmt);
  fmtsizeIn = fmtsizeof(fmtIn);
  if (sizIn > 0 && datIn == NULL) return argument_list_error;
  if (strlen(eqm) > EQM_NAME_SIZE || strlen(prp) > PROPERTY_NAME_SIZE) return string_too_long;
  if ((dsize = siz*fmtsize) < 0) return out_of_range;
  if ((dsizeIn = sizIn*fmtsizeIn) < 0) return out_of_range;
  if ((2 + PRODUCER_SIZE + dsize) > IPX_BUFFER_SIZE) return use_stream_transport;
  if ((HEADERSTRUCT_SIZE + SUBSCRIPTION_SIZE + dsizeIn) > IPX_BUFFER_SIZE) return use_stream_transport;
  if (timout < 1) timout = 1;
  if (gTineServicesProtocol < 5 || gTineServicesProtocol > 6) return illegal_protocol;

pRETRY:
  memset(&c,0,sizeof(ConnTableEntry));
  c.mode = CM_SINGLE; 
  c.starttime = (UINT32)time(NULL);
  strncpy(c.prpName,prp,PROPERTY_NAME_SIZE);
  strncpy(c.devName,dev,DEVICE_NAME_SIZE);
  if (stagOut != NULL) strncpy(c.strTagOut,stagOut,TAG_NAME_SIZE);
  if (stagIn != NULL) strncpy(c.strTagIn,stagIn,TAG_NAME_SIZE);
  c.fecIdx = fecindex;
  c.sizeIn = sizIn; 
  c.formatIn = fmtIn;
  c.sizeOut = siz; 
  c.formatOut = fmt;
  c.access = access;
  c.datamtu = SUBMTU(FecTbl[fecindex].TransportProtocol,gTineServicesProtocol); /* input data mtu */ 
  c.blkidIn = 0;
  c.blknumIn = 1;
  c.numblksIn = 1;

  sstepsize = SUB_STEPSIZE(gTineServicesProtocol);
  pstepsize = PHDR_STEPSIZE(gTineServicesProtocol);

  AssignOutgoingHeader(h,lclWorkArea);

tRETRY:
  c.tineProtocol = gTineServicesProtocol;
  FecTbl[fecindex].TineProtocol = gTineServicesProtocol;
  /* below assumes that a data payload has already been prepared (swapped prior to entering) */
  if (sizIn > 0)
  {
    memcpy(&lclWorkArea[HEADERSTRUCT_SIZE+sstepsize],datIn,dsizeIn);
  }
  prepOutgoingSubscription(gTineServicesProtocol,&c,eqm,dsizeIn,tid,&lclWorkArea[HEADERSTRUCT_SIZE]);

  h->TotalSize = HEADERSTRUCT_SIZE + sstepsize + dsizeIn;
  memcpy(h->UserName,gUserName,USERNAME_SIZE);
  h->Protocol = gTineServicesProtocol;
  h->EqmSize = 1;
  PrepOutgoingHeader(h,lclWorkArea);

  switch (cc = tCall(fecindex,lclWorkArea,timout))
  {
    case 0: 
      break;
    case illegal_protocol:
      if (gTineServicesProtocol == 6) 
      {
        gTineServicesProtocol = 5;
        goto pRETRY;
      }
    case link_not_open:
      if (--retry > 0) goto tRETRY;
    default:
      ccerr(cc);
  }

  prepIncomingSubscription(gTineServicesProtocol,pPHdr,&lclWorkArea[2]);
  if (pPHdr->SubscriptionID != tid) ccerr(connection_timeout);
  if ((cc=pPHdr->CompletionCode) == 0 || (cc & CE_SENDDATA))
  {
    if (dsize < (dretsiz=pPHdr->msgsize-pstepsize)) ccerr(illegal_format);
    memcpy(datOut,&lclWorkArea[2+pstepsize],dretsiz);
  }
  else
  {
    strncpy(tExecLinkReturnedCompletion,(char *)&lclWorkArea[2+pstepsize],STSSIZE(gTineServicesProtocol));
  }
err:
  tid = (tid + 1) % 10000;
  if (cc && tineDebug)
  {
    dbglog("trivial ExecLink error: <%d> %s",cc, cc > 0 && cc < numErr ? erlst[cc] : "unknown");
  }
  tExecLinkReturnedDataSize = fmtsize > 0 ? dretsiz / fmtsize : 0;
  return cc;
}
