#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "srvcore.h"
#include "swaplib.h"
#include "globals.h"
#include "dbtools.h"
#include "srvip.h"
#include "toolkit.h"
#include "ttycmd.h"
#include "wincelib.h"

int kbhit(void)
{
  return 0; 
}
int getch(void)
{
  int c = getchar();
  putchar(c);
  return c;
}
int putch(int c)
{
  return putchar(c);
}
int perror(const char *errstr)
{
  printf("error : %s",errstr);
  return 0;
}

/***** TIME WINCE functions ****/
// Many but not all of the functions were created from the following source:
// Copyright (C) 2002 Michael Ringgaard. All rights reserved.
// do we need this ?

// Defines 
#define TIME_FAILURE	0xFFFFFFFF
#define ASC_BUFF_SIZE	26  // Ascii buffer size is 26 bytes, (24 chars and CR+LF)
#define SEC_IN_HOUR     3600L
#define SECS_IN_MIN		60L
#define DAYSPERWEEK		7
#define YEAR0           1900
#define EPOCH_YR        1970
#define SECS_DAY        (24L * 60L * 60L)
#define LEAPYEAR(year)  (!((year) % 4) && (((year) % 100) || !((year) % 400)))
#define TIME_MAX        2147483647L

// Local variables
// Number of seconds between local time and UTC time, includes DST bias
//
LONG _localtime;

// Is the local time in daylight savings time
//
DWORD _isdst;

// Bias for daylight savings time
//
int _dstBias;

// Contains the time zone string
//
char tz_name[2][32];

// Contains the 1/1/1970 reference date/time
//
const SYSTEMTIME st1970 = {1970, 1,	4, 1, 0, 0, 0, 0};

// Contains the number of days per month for 
// non leap and leap years
//
const int _ytab[2][12] = 
{
  {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
  {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
};
static char *aday[] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

// Contains the days of the week full name
//
static char *day[] = {
    "Sunday", "Monday", "Tuesday", "Wednesday",
    "Thursday", "Friday", "Saturday"
};

// Contains the months of the year abreviation
//
static char *amonth[] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

// Contains the months of the year full name
//
static char *month[] = {
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
};

// Get the JulianDay from a Gregorian Date for number of days into the current
// year
// Algorithm from: http://www.vsg.cape.com/~pbaum/date/jdalg2.htm
//
DWORD JulianDays(SYSTEMTIME * st)
{
	int m = 0;
	int y = 0;
	double jdd = 0;
	double jddYearStart = 0;

	// Calculate the Julian day for the beginning of the year
	//
    m = 13;
	y = st->wYear - 1;
    jddYearStart = 1 + (153 * m - 457) / 5 + 365 * y + (y / 4) - (y / 100) + (y / 400) + 1721118.5;

	// Calculate Julian Day for Current Date
	//
	if (st->wMonth >= 3)
	{
		m = st->wMonth;
		y = st->wYear;
	}
    jdd = st->wDay + (153 * m - 457) / 5 + 365 * y + (y / 4) - (y / 100) + (y / 400) + 1721118.5;

	// Subract the year start Julian date from the Current Julian date to get
	// the number of Julian days from the year start
	//
	return (DWORD)(jdd - jddYearStart);
}


LONGLONG SystemTimeToSecondsSince(SYSTEMTIME *st)
{
	ULARGE_INTEGER	uli;
	FILETIME		ft;
	ULARGE_INTEGER	uli1970;
	FILETIME		ft1970;

	// convert to a FILETIME
	// Gives number of 100-nanosecond intervals since January 1, 1601 (UTC)
	//
	if(!SystemTimeToFileTime(st, &ft))
	{
		return TIME_FAILURE;
	}

	// convert to a FILETIME
	// Gives number of 100-nanosecond intervals since January 1, 1970 (UTC)
	//
	if(!SystemTimeToFileTime(&st1970, &ft1970))
	{
		return TIME_FAILURE;
	}

	// Copy file time structures into ularge integer so we can do
	// the math more easily
	//
	memcpy(&uli, &ft, sizeof(uli));
	memcpy(&uli1970, &ft1970, sizeof(uli1970));

	// Subtract the 1970 number of 100 ns value from the 1601 100 ns value
	// so we can get the number of 100 ns value between 1970 and now
	// then devide be 10,000,000 to get the number of seconds since 1970
	//
	uli.QuadPart = ((uli.QuadPart - uli1970.QuadPart) / 10000000);

	return (LONGLONG)uli.QuadPart;
}

#define IS_LEAP(yr) (((yr)%4==0 || (yr)%100==0) && !(yr)%400==0)
static char _day_mo_leap[12] = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
static char _day_mo[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

struct tm * __cdecl localtime ( const time_t *ptime )
{
	static struct tm ptm;
	int i;
	if( !ptime ) return NULL;

	_t_cnv.t_val = *ptime;

    ptm.tm_sec		= _t_cnv.tm_val.ss ;     /* seconds after the minute - [0,59] */
    ptm.tm_min		= _t_cnv.tm_val.mm;     /* minutes after the hour - [0,59] */
    ptm.tm_hour	= _t_cnv.tm_val.hh;    /* hours since midnight - [0,23] */
    ptm.tm_mday	= _t_cnv.tm_val.dd;    /* day of the month - [1,31] */
    ptm.tm_mon		= _t_cnv.tm_val.mo-1;     /* months since January - [0,11] */
    ptm.tm_year	= _t_cnv.tm_val.yr;    /* years since 1900 */
    ptm.tm_wday	= _t_cnv.tm_val.wd;    /* days since Sunday - [0,6] */

	ptm.tm_yday = _t_cnv.tm_val.dd;    /* days since January 1 - [0,365] */
	for( i=0; i<ptm.tm_mon; i++ )
		ptm.tm_yday += IS_LEAP(_t_cnv.tm_val.yr+1900)?_day_mo_leap[i] : _day_mo[i] ; 

	ptm.tm_isdst = 0;   /* daylight savings time flag  - NOT IMPLEMENTED */
	return &ptm;
}

time_t __cdecl time(time_t * timeptr)
{
	SYSTEMTIME stm;
	GetLocalTime(&stm);

	_t_cnv.tm_val.yr = stm.wYear-1900; 
	_t_cnv.tm_val.mo = stm.wMonth; 
	_t_cnv.tm_val.dd = stm.wDay; 
	_t_cnv.tm_val.hh = stm.wHour; 
	_t_cnv.tm_val.mm = stm.wMinute; 
	_t_cnv.tm_val.ss = stm.wSecond; 
	_t_cnv.tm_val.wd = stm.wDayOfWeek; 

	if( timeptr) 
		*timeptr = _t_cnv.t_val;
	return _t_cnv.t_val;
}

static void strfmt(char *str, const char *fmt, ...)
{
	int ival, ilen;
	char *sval;
	static int pow[5] = { 1, 10, 100, 1000, 10000 };
	va_list vp;

	va_start(vp, fmt);
	while (*fmt)
	{
		if (*fmt++ == '%')
		{
			ilen = *fmt++ - '0';
			if (ilen == 0)                // zero means string arg
			{
				sval = va_arg(vp, char*);
				while (*sval)
					*str++ = *sval++;
			}
			else                          // always leading zeros
			{
				ival = va_arg(vp, int);
				while (ilen)
				{
					ival %= pow[ilen--];
					*str++ = (char)('0' + ival / pow[ilen]);
				}
			}
		}
		else  *str++ = fmt[-1];
	}
	*str = '\0';
	va_end(vp);
}


size_t strftime(char *s, size_t maxs, const char *f, const struct tm *t)
{
	int w;
	char *p, *q, *r;
	static char	buf[26];

	p = s;
	q = s + maxs - 1;
	while ((*f != '\0'))
	{
	  if (*f++ == '%')
	  {
	 	r = buf;
		switch (*f++)
		{
			case '%' :
				r = "%";
				break;

			case 'a' :
				r = aday[t->tm_wday];
				break;

			case 'A' :
				r = day[t->tm_wday];
				break;

			case 'b' :
				r = amonth[t->tm_mon];
				break;

			case 'B' :
				r = month[t->tm_mon];
				break;

			case 'c' :
				strfmt(r, "%0 %0 %2 %2:%2:%2 %4",
					  aday[t->tm_wday], amonth[t->tm_mon],
					  t->tm_mday,t->tm_hour, t->tm_min,
					  t->tm_sec, t->tm_year+1900);
				break;

			case 'd' :
				strfmt(r,"%2",t->tm_mday);
				break;

			case 'H' :
				strfmt(r,"%2",t->tm_hour);
				break;

			case 'I' :
				strfmt(r,"%2",(t->tm_hour%12)?t->tm_hour%12:12);
				break;

			case 'j' :
				strfmt(r,"%3",t->tm_yday+1);
				break;

			case 'm' :
				strfmt(r,"%2",t->tm_mon+1);
				break;

			case 'M' :
				strfmt(r,"%2",t->tm_min);
				break;

			case 'p' :
				r = (t->tm_hour>11)?"PM":"AM";
				break;

			case 'S' :
				strfmt(r,"%2",t->tm_sec);
				break;

			case 'U' :
				w = t->tm_yday/7;
				if (t->tm_yday%7 > t->tm_wday)
			  	  w++;
				strfmt(r, "%2", w);
				break;

			case 'W' :
				w = (t->tm_yday + DAYSPERWEEK -
					(t->tm_wday ?
					(t->tm_wday - 1) :
					(DAYSPERWEEK - 1))) / DAYSPERWEEK;
				strfmt(r, "%2", w);
				break;

			case 'w' :
				strfmt(r,"%1",t->tm_wday);
				break;

			case 'x' :
				strfmt(r, "%2/%2/%2", t->tm_mon + 1,
					   t->tm_mday, t->tm_year+1900);
				break;

			case 'X' :
				strfmt(r, "%2:%2:%2", t->tm_hour,
					   t->tm_min, t->tm_sec);
				break;

			case 'y' :
				strfmt(r,"%2",t->tm_year%100);
				break;

			case 'Y' :
				strfmt(r,"%4",t->tm_year+1900);
				break;

			case 'Z' :
				r = (t->tm_isdst && tz_name[1][0])?tz_name[1]:tz_name[0];
				break;

			default:
				buf[0] = '%';		// reconstruct the format
				buf[1] = f[-1];
				buf[2] = '\0';
				if (buf[1] == 0)
			  	  f--;			// back up if at end of string
		}
		while (*r)
		{
		  if (p == q)
		  {
			*q = '\0';
			return 0;
		  }
		  *p++ = *r++;
		}
	}
	else
	{
	  if (p == q)
	  {
		*q = '\0';
		return 0;
	  }
	  *p++ = f[-1];
	}
  }
  *p = '\0';
  return p - s;
}

char *asctime(const struct tm* tmptr)
{
	static char ascbuf[ASC_BUFF_SIZE];
    strftime(ascbuf, ASC_BUFF_SIZE, "%c\n", tmptr);
    return ascbuf;
}

char *ctime(const time_t *utc)
{
  return asctime(localtime(utc));
}

void TmToSystemTime(struct tm * tmbuffer, SYSTEMTIME *st)
{
  st->wHour = tmbuffer->tm_hour;
  st->wDay = tmbuffer->tm_mday;
  st->wMinute = tmbuffer->tm_min;
  st->wMonth = tmbuffer->tm_mon + 1;
  st->wSecond = tmbuffer->tm_sec;
  st->wDayOfWeek = tmbuffer->tm_wday;
  st->wYear = tmbuffer->tm_year + 1900;
  st->wMilliseconds = 0;
}

void SystemTimeToTm(SYSTEMTIME *st, struct tm * tmbuffer)
{
  tmbuffer->tm_hour  = st->wHour;
  tmbuffer->tm_mday  = st->wDay;
  tmbuffer->tm_min   = st->wMinute;
  tmbuffer->tm_mon   = st->wMonth - 1;
  tmbuffer->tm_sec   = st->wSecond;
  tmbuffer->tm_wday  = st->wDayOfWeek;
  tmbuffer->tm_yday  = JulianDays(st);		// Julian days, numer of days since Jan 1
  tmbuffer->tm_year  = st->wYear - 1900;
  tmbuffer->tm_isdst = _isdst;				// Is Daylight Savings Time
}

/* Set the time zone information needed for the rest of the methods */
void SetTz(SYSTEMTIME *_st)
{
	TIME_ZONE_INFORMATION	tz;
	SYSTEMTIME				st = *_st;
	FILETIME				ftDT; // Daylight savings file time
	FILETIME				ftST; // Standard time file time
	FILETIME				ft;   // file time to compare
	int						i;

	GetTimeZoneInformation(&tz);

	// Convert current system time, daylight savings changover time
	// and standard time to file time to see if current time is between
	// the two dates.  If so then we are in daylight savings otherwise we
	// are not.
	//
	SystemTimeToFileTime(&st, &ft);
	tz.DaylightDate.wYear = st.wYear;
	tz.StandardDate.wYear = st.wYear;
  SystemTimeToFileTime(&tz.DaylightDate, &ftDT);
	SystemTimeToFileTime(&tz.StandardDate, &ftST);

  // -1 First file time is earlier than second file time. 
  //  0 First file time is equal to second file time. 
  //  1 First file time is later than second file time. 
	//
	if ((CompareFileTime(&ft,&ftDT) >= 0) && (CompareFileTime(&ft,&ftST) <= 0) )
	{
		_isdst = TRUE;
	}
	else
	{
		_isdst = FALSE;
	}

	// Set localtime difference in seconds from UTC
	//
	if (_isdst)
	{
		// Adjust for Daylight Savings Time and convert to seconds
		//
		_localtime = (tz.Bias + tz.DaylightBias) * SECS_IN_MIN;
	}
	else
	{
		// Convert to seconds
		//
		_localtime = tz.Bias * SECS_IN_MIN;
	}

	_dstBias = tz.DaylightBias;

	// Set the standard and daylight strings
	//
	for (i=0;i<32;i++)
	{
		tz_name[0][i] = (char)tz.StandardName[i];
		tz_name[1][i] = (char)tz.DaylightName[i];
	}
}

time_t mktime(struct tm *timeptr)
{
  SYSTEMTIME st;
  DWORD day;
  DWORD year;
  DWORD seconds; 
  DWORD overflow;
  DWORD tm_year;
  DWORD yday, month;
  struct tm * tptr=NULL;

  TmToSystemTime(tptr, &st);
  SetTz(&st);
  
  day = 0;
  year = 0;
  seconds = 0;
  
  // see if seconds are < 0
  while(tptr->tm_sec < 0)
  {
	// steal 60 seconds from the minutes
	tptr->tm_sec += 60;
	tptr->tm_min--;
  }
  // roll any seconds > 60 into the minutes
  tptr->tm_min += tptr->tm_sec / 60;
  // then crop them off
  tptr->tm_sec %= 60;

  // see if minutes are < 0
  while(tptr->tm_min < 0)
  {
 	// steal 60 minutes from the hours
    tptr->tm_min += 60;
	tptr->tm_hour--;
  }
  // roll any minutes > 60 into the hours
  tptr->tm_hour += tptr->tm_min / 60;
  // then crop them off
  tptr->tm_min %= 60;

  // see if hours are < 0
  while(tptr->tm_hour < 0)
  {
	// steal 24 hours from the days
	tptr->tm_hour += 24;
	day--;
  }
  // keep any "excess" days (tm doesn't have a convenient place for this)
  day += tptr->tm_hour / 24;
  // crop
  tptr->tm_hour %= 24;

  // roll any months > 12 into the years
  tptr->tm_year += tptr->tm_mon / 12;
  // then crop the off
  tptr->tm_mon %= 12;

  // see if months are < 0
  if (tptr->tm_mon < 0) 
  {
 	// steal 12 months from the years
	tptr->tm_mon += 12;
	tptr->tm_year--;
  }

  // add number of days into the month to total day
  day += (tptr->tm_mday - 1);
 
  // if days are < 0 then calculate the number of days
  // checking to see if the month is a leap year month
  while (day < 0) 
  {
	// If months are < 0 then steal 12 months from number of years
	// for the day calculation
	if(--tptr->tm_mon < 0) 
	{
   	  tptr->tm_year--;
	  tptr->tm_mon = 11;
	}
	day += _ytab[LEAPYEAR(YEAR0 + tptr->tm_year)][tptr->tm_mon];
  }
  // if day is greater then the number of days in the month
  // subtract the number of days in the month and adjust the 
  // month
  while (day >= (ULONG)_ytab[LEAPYEAR(YEAR0 + tptr->tm_year)][tptr->tm_mon]) 
  {
	day -= _ytab[LEAPYEAR(YEAR0 + tptr->tm_year)][tptr->tm_mon];
	if (++(tptr->tm_mon) == 12) 
	{
  	  tptr->tm_mon = 0;
	  tptr->tm_year++;
	}
  }
  tptr->tm_mday = day + 1;
  year = EPOCH_YR;

  // if year is less then 1970 then return error
  if ((ULONG)tptr->tm_year < year - YEAR0) return (time_t) -1;

  seconds = 0;
  day = 0;                      // Means days since day 0 now
  overflow = 0;

  // Assume that when day becomes negative, there will certainly
  // be overflow on seconds.
  // The check for overflow needs not to be done for leapyears
  // divisible by 400.
  // The code only works when year (1970) is not a leapyear.
  tm_year = tptr->tm_year + YEAR0;

  // make sure we are not past the max year for 32-bit number
  if (TIME_MAX / 365 < tm_year - year) overflow++;

  // calculate number of days since EPOCH
  day = (tm_year - year) * 365;

  if (TIME_MAX - day < (tm_year - year) / 4 + 1) overflow++;

  day += (tm_year - year) / 4 + ((tm_year % 4) && tm_year % 4 < year % 4);
  day -= (tm_year - year) / 100 + ((tm_year % 100) && tm_year % 100 < year % 100);
  day += (tm_year - year) / 400 + ((tm_year % 400) && tm_year % 400 < year % 400);

  // setup for calculation of the yday or Julian day since Jan 1
  yday = month = 0;

  // add up the number of days for the preceding months
  while (month < (ULONG)tptr->tm_mon)
  {
	yday += _ytab[LEAPYEAR(tm_year)][month];
	month++;
  }
  // add the number of days in the current month
  yday += (tptr->tm_mday - 1);

  // make sure the didn't overflow
  if (day + yday < 0) overflow++;

  day += yday;

  // set the year day in the structure
  tptr->tm_yday = yday;

  // calculate the weekday
  tptr->tm_wday = (day + 4) % 7;               // Day 0 was thursday (4)

  // start the seconds calculation by totaling the hours, min, seconds
  seconds = ((tptr->tm_hour * 60L) + tptr->tm_min) * 60L + tptr->tm_sec;

  // make sure we are not going to overflow
  if ((TIME_MAX - seconds) / SECS_DAY < day) overflow++;

  // calculate the number of seconds for the number of days
  seconds += day * SECS_DAY;

  // Now adjust according to timezone and daylight saving time
  if (((_localtime > 0) && (TIME_MAX - (ULONG)_localtime < seconds))
     || ((_localtime < 0) && ((LONG)seconds < - _localtime)))
  	  overflow++;

  // Adjust for local time zone
  seconds += _localtime;

  // return error if we are going to blow the max values
  if (overflow) return (time_t) -1;

  if ((time_t) seconds != seconds) return (time_t) -1;

  // return the number of seconds since EPOCH
  return (time_t) seconds;  
}

int dayOfYear(int year, int month, int day)
{
  int totalDay = 0;
  int i = 0;
  
  if IS_LEAP(year)
    for (i=0;i<month;i++)  totalDay += _day_mo[i];
  else
    for (i=0;i<month;i++)  totalDay += _day_mo_leap[i];

  totalDay += day;
  
  return (totalDay);
}


int ftime(struct timeb *tb)
{
  SYSTEMTIME stm;
  struct tm *tmbuf = NULL;
  LONGLONG secs;
  
  GetLocalTime(&stm);
   
  secs = SystemTimeToSecondsSince(&stm);
  tb->time    = secs;
  tb->millitm = stm.wMilliseconds; 
  
  return 0;
}
void tzset(void)
{ 
  // not used in CE ?
  //tzname[0] = "std";
  //tzname[1] = "dst";
  return;
}

char *getenv(const char *envstr)
{
  // STD_TIME_STR
  // DST_TIME_STR
  // CONTROLUSER
  // return (char *)envstr;
  return NULL;
}
char *putenv(const char *envstr)
{
  return NULL;
}
char *strupr(char *s)
{
  char *c;
  for (c=s; c && *c; c++) if (*c >= 'a' && *c <= 'z') *c -= 32;
  return s;
}
char *strlwr(char *s)
{
  char *c;
  for (c=s; c && *c; c++) if (*c >= 'A' && *c <= 'Z') *c += 32;
  return s;
}



/*------------------------------------------------------------------------------*/

/* __io.h__ */
/* Hack io.h function with stdio.h functions */
/* ASSUMPTION : int can hold FILE* */
static TCHAR _nh_cwd[MAX_PATH];
const int MAGIC_OFFSET=5;
#define FILE_TABLE_SIZE  256
static HANDLE _nh_file_table[FILE_TABLE_SIZE];
static int file_pointer = -1;

static HANDLE get_file_handle(int i) 
{
	i -= MAGIC_OFFSET;
	if( i>=0 && i<FILE_TABLE_SIZE )
		return _nh_file_table[i];
	else 
		return INVALID_HANDLE_VALUE;
} 

long __cdecl lseek( int f, long offset, int origin )
{
	HANDLE p = get_file_handle(f);
	DWORD fpos;
	switch(origin) {
	case SEEK_SET:
		fpos = SetFilePointer(p, offset, NULL, FILE_BEGIN);
		break;
	case SEEK_CUR:
		fpos = SetFilePointer(p, offset, NULL, FILE_CURRENT);
		break;
	case SEEK_END:
		fpos = SetFilePointer(p, offset, NULL, FILE_END);
		break;
	default:
		fpos = 0xFFFFFFFF;
		break;
	}
	if( fpos==0xFFFFFFFF ) return -1;
	else return (long)fpos;
}

int __cdecl read( int f, void *buffer, unsigned int count )
{
	HANDLE p = get_file_handle(f);
	DWORD bytes_read;
	if( !ReadFile(p, buffer, count, &bytes_read, NULL) )
		return -1;
	else 
		return (int)bytes_read;
}

int __cdecl write( int f, const void *buffer, unsigned int count )
{
	HANDLE p = get_file_handle(f);
	DWORD bytes_written;
	if( !WriteFile(p, buffer, count, &bytes_written, NULL) ) 
		return -1;
	else 
		return (int)bytes_written;
}

int __cdecl close(int f) {
	int retval;
	f -= MAGIC_OFFSET;
	if( f<0 || f>=FILE_TABLE_SIZE )	return -1;
	retval = (CloseHandle(_nh_file_table[f])? 0 : -1);
	_nh_file_table[f] = INVALID_HANDLE_VALUE;
	return retval;
}

int stricmp(char *s1,char *s2)
{
  char *c1,*c2;
  for (c1=s1,c2=s2; *c1 != 0 && *c2 != 0; c1++,c2++)
    if (toupper(*c1) != toupper(*c2)) break;
  return toupper(*c2) - toupper(*c1);
}

int strnicmp(char *s1,char *s2,int n)
{
  char *c1,*c2;
  int i;
  for (c1=s1,c2=s2,i=0; *c1 != 0 && *c2 != 0; c1++,c2++,i++)
  {
    if (toupper(*c1) != toupper(*c2)) break;
    if (i >= n-1) break;
  }
  return toupper(*c2) - toupper(*c1);
}

char *strrev(char *s)
{
  char c;
  register int i,len = (int)strlen(s);
  for (i=0; i<len/2; i++) { c = s[i]; s[i] = s[len-1-i]; s[len-1-i] = c; }
  return s;
}


/* Functions used in CDI */
int remove(const char *filename)
{
  return 0;
}

int rename ( const char * oldname, const char * newname )
{
  MoveFile( (LPCTSTR) oldname, (LPCTSTR) newname );
  return 0;
}
